from functools import partial
from hashlib import blake2b
from typing import Iterable, Any, Generator

from ..cache.disk.cache import key_to_digest
from .transform import TransformContainer
from ..engine.base import Node, Request, Response, Command
from ..engine.edges import StaticHash, StaticGraph
from ..engine.node_hash import NodeHashes, NodeHash


class HashDigestEdge(StaticGraph, StaticHash):
    def __init__(self):
        super().__init__(arity=1)
        # TODO: find a way to pass different hashers
        self._hasher = partial(blake2b, digest_size=64)

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]

    def evaluate(self) -> Generator[Request, Response, Any]:
        value = yield Command.ParentValue, 0
        output = yield Command.CurrentHash,

        pickled, digest = key_to_digest(self._hasher, output.value)
        return value, output.value, digest, pickled


class HashDigestContainer(TransformContainer):
    def __init__(self, names: Iterable[str]):
        inputs, outputs, edges = [], [], []
        for name in names:
            inp, out = Node(name), Node(name)
            inputs.append(inp)
            outputs.append(out)
            edges.append(HashDigestEdge().bind(inp, out))

        super().__init__(inputs, outputs, edges, forward_virtual=True, backward_virtual=True)



import logging
from typing import Any, Tuple, Sequence

from .index import CacheIndexStorage
from ...engine import NodeHash
from ..base import Cache
from ..pickler import dumps, PREVIOUS_VERSIONS
from ...storage.interface import Key, RemoteStorage, HashStorage

logger = logging.getLogger(__name__)


class DiskCache(Cache):
    def __init__(self, local: Sequence[CacheIndexStorage], remote: Sequence[RemoteStorage], fetch: bool):
        super().__init__()
        self.cache = HashStorage(local, remote)
        self.algorithm = self.cache.hash.build()
        self.fetch = fetch

    def prepare(self, param: NodeHash) -> Tuple[Key, Any]:
        raw = param.value
        pickled, key = key_to_digest(self.algorithm, raw)
        # OPTIMIZATION: if PREVIOUS_VERSIONS is empty can return just the pickled part
        return key, (raw, pickled)

    def get(self, key: Key, context) -> Tuple[Any, bool]:
        raw, pickled = context
        logger.info('Looking for key %s', key)

        # try to load
        value, exists = self.cache.read(key, pickled, fetch=self.fetch)
        if exists:
            logger.info('Key %s found', key)
            return value, exists

        # the cache is empty, but we can try and restore it from an older version
        for version in reversed(PREVIOUS_VERSIONS):
            local_pickled, local_digest = key_to_digest(self.algorithm, raw, version)

            # we can simply load the previous version, because nothing really changed
            value, exists = self.cache.read(local_digest, local_pickled, fetch=self.fetch)
            if exists:
                logger.info('Key %s found in previous version (%d). Updating', key, version)
                # and store it for faster access next time
                self.cache.write(key, value, pickled)
                return value, exists

        logger.info('Key %s not found', key)
        return None, False

    def set(self, key: Key, value: Any, context):
        raw, pickled = context
        logger.info('Saving key %s', key)
        self.cache.write(key, value, pickled)


def key_to_digest(algorithm, key, version=None):
    pickled = dumps(key, version=version)
    digest = algorithm(pickled).hexdigest()
    return pickled, digest



import sys
from abc import ABC, abstractmethod
from multiprocessing.pool import ThreadPool

from .base import Command

if sys.version_info[:2] < (3, 7):
    from queue import Queue as SafeQueue
else:
    from queue import SimpleQueue as SafeQueue


class Thunk:
    def __init__(self, parent):
        self.parent = parent
        self.ready = False
        self.error = self.value = None


class Frame(Thunk):
    def __init__(self, stack, commands, parent):
        super().__init__(parent)
        self.stack = stack
        self.commands = commands


class Executor(ABC):
    def __init__(self, frame: Frame):
        self.frame = frame

    def push(self, x):
        self.frame.stack.append(x)

    def pop(self):
        return self.frame.stack.pop()

    def peek(self):
        return self.frame.stack[-1]

    def push_command(self, x):
        self.frame.commands.append(x)

    def pop_command(self):
        return self.frame.commands.pop()

    @abstractmethod
    def next_frame(self):
        pass

    @abstractmethod
    def enqueue_frame(self, x):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def call(self, func, args, kwargs):
        pass


class Backend:
    @abstractmethod
    def build(self, frame: Frame) -> Executor:
        pass


class Synchronous(Backend):
    class _Executor(Executor):
        def __init__(self, frame: Frame):
            super().__init__(frame)
            self.frames = SafeQueue()

        def clear(self):
            q = self.frames
            while not q.empty():
                assert q.get_nowait().ready

        def next_frame(self):
            self.frame = self.frames.get_nowait()

        def enqueue_frame(self, x):
            self.frames.put_nowait(x)

        def call(self, func, args, kwargs):
            self.push(func(*args, **kwargs))

    def build(self, frame: Frame) -> Executor:
        return self._Executor(frame)


class Threads(Backend):
    class _Executor(Executor):
        def __init__(self, frame: Frame, queue: SafeQueue):
            super().__init__(frame)
            self.frames, self.requests = SafeQueue(), queue

        def clear(self):
            q = self.frames
            while not q.empty():
                assert q.get_nowait().ready

        def next_frame(self):
            self.frame = self.frames.get()

        def enqueue_frame(self, x):
            self.frames.put(x)

        def call(self, func, args, kwargs):
            thunk = Thunk(self.frame)
            self.requests.put((func, args, kwargs, thunk, self.frames))
            self.push_command((Command.AwaitThunk, thunk))
            self.next_frame()

    def build(self, frame: Frame) -> Executor:
        return self._Executor(frame, self.thunks)

    def __init__(self, n: int):
        self.n = n
        self.thunks = SafeQueue()
        self.pool = ThreadPool(n, self._loop, (self.thunks,))
        self.pool.close()

    @staticmethod
    def _loop(thunks: SafeQueue):
        while True:
            value = thunks.get()
            if value is None:
                break

            func, args, kwargs, thunk, frames = value
            try:
                thunk.value = func(*args, **kwargs)
            except BaseException as e:
                thunk.error = e

            thunk.ready = True
            frames.put(thunk.parent)

    def __del__(self):
        for _ in range(self.n):
            self.thunks.put(None)
        self.pool.join()


DefaultBackend = Synchronous()



import shutil
import socket
import tempfile
from pathlib import Path
from typing import Union, Sequence, Callable, Any, Tuple

import paramiko
from paramiko import SSHClient, AuthenticationException, SSHException
from paramiko.config import SSH_PORT, SSHConfig
from scp import SCPClient, SCPException

from .config import load_config, HashConfig
from .digest import key_to_relative
from .interface import RemoteStorage, Key
from ..utils import PathLike


class UnknownHostException(SSHException):
    pass


class SSHLocation(RemoteStorage):
    def __init__(self, hostname: str, root: PathLike, port: int = SSH_PORT, username: str = None, password: str = None,
                 key: Union[Path, Sequence[Path]] = ()):
        ssh = SSHClient()
        ssh.load_system_host_keys()
        # TODO: not safe
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if isinstance(key, Path):
            key = str(key)

        config_path = Path('~/.ssh/config').expanduser()
        if config_path.exists():
            with open(config_path) as f:
                config = SSHConfig()
                config.parse(f)
                host = config.lookup(hostname)

                hostname = host.get('hostname', hostname)
                port = host.get('port', port)
                username = host.get('user', username)
                key = host.get('identityfile', key)

        self.hostname, self.port, self.username, self.password, self.key = hostname, port, username, password, key
        self.root = Path(root)
        self.ssh = ssh
        self.levels = self.hash = None

    def fetch(self, keys: Sequence[Key], store: Callable[[Key, Path], Any],
              config: HashConfig) -> Sequence[Tuple[Any, bool]]:

        results = []
        try:
            self.ssh.connect(
                self.hostname, self.port, self.username, self.password, key_filename=self.key,
                auth_timeout=10
            )
        except AuthenticationException:
            raise AuthenticationException(self.hostname) from None
        except socket.gaierror:
            raise UnknownHostException(self.hostname) from None

        try:
            with SCPClient(self.ssh.get_transport()) as scp, tempfile.TemporaryDirectory() as temp_dir:
                source = Path(temp_dir) / 'source'

                for key in keys:
                    try:
                        self._get_config(scp, config)
                        scp.get(str(self.root / key_to_relative(key, self.levels)), str(source), recursive=True)
                        value = store(key, source)
                        shutil.rmtree(source)

                        results.append((value, True))
                    except (SCPException, socket.timeout):
                        results.append((None, False))

        finally:
            self.ssh.close()

        return results

    def _get_config(self, scp, config):
        if self.levels is None:
            with tempfile.TemporaryDirectory() as temp_dir:
                temp = Path(temp_dir) / 'config.yml'
                scp.get(str(self.root / 'config.yml'), str(temp))
                remote_config = load_config(temp_dir)
                self.hash, self.levels = remote_config.hash, remote_config.levels

        assert self.hash == config, (self.hash, config)



class GraphError(Exception):
    pass


class DependencyError(GraphError):
    pass


class FieldError(Exception):
    pass


class StorageCorruption(OSError):
    """
    Denotes various problems with disk-based storage or persistent cache
    """



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Tuple, Callable, Any, Sequence, Union

from tqdm import tqdm

from .config import HashConfig

Key = str
logger = logging.getLogger(__name__)


class StorageError(Exception):
    pass


class WriteError(StorageError):
    pass


class ReadError(StorageError):
    pass


# TODO: deprecated
StoreError = WriteError
QueryError = ReadError


class LocalStorage(ABC):
    def __init__(self, hash: HashConfig, levels: Sequence[int]):
        self.hash = hash
        self.algorithm = self.hash.build()
        self.levels = levels

    @abstractmethod
    def write(self, key, value: Any, context: Any) -> bool:
        """
        Write a ``value`` to a given ``key``.
        Returns True if the value was written (or already present).
        """

    @abstractmethod
    def read(self, key, context: Any) -> Tuple[Any, bool]:
        """
        Read the value given the ``key``.
        Returns a pair (value, success).
        If success is False - the value could not be read.
        """

    @abstractmethod
    def contains(self, key, context: Any) -> bool:
        """
        Returns whether the ``key`` is present in the storage.
        """

    @abstractmethod
    def replicate(self, key, source: Path, context: Any) -> bool:
        """
        Populates the storage at ``key`` from a ``source`` path.
        """


class RemoteStorage:
    hash: HashConfig
    levels: Sequence[int]

    @abstractmethod
    def fetch(self, keys: Sequence[Key], store: Callable[[str, Path], Any],
              config: HashConfig) -> Sequence[Tuple[Any, bool]]:
        """
        Fetches the value for ``key`` from a remote location.
        """


class HashStorage:
    def __init__(self, local: Sequence[LocalStorage], remote: Sequence[RemoteStorage] = ()):
        if not local:
            raise ValueError('The storage must have at least 1 local config')

        reference = local[0].hash
        for loc in local[1:]:
            if loc.hash != reference:
                raise ValueError('Local storage locations have inconsistent hash algorithms')

        self.local, self.remote, self.hash = local, remote, reference

    def write(self, key, value, context) -> bool:
        for local in self.local:
            if local.write(key, value, context):
                return True

        return False

    def read(self, key, context, *, fetch: bool) -> Tuple[Any, Union[None, bool]]:
        # try to find locally
        for local in self.local:
            value, success = local.read(key, context)
            if success:
                return value, True

        # try to fetch from remote
        status = False
        if fetch:
            for remote in self.remote:
                (local, success), = remote.fetch([key], lambda k, base: self._replicate(k, base, context), self.hash)
                if success:
                    if local is WriteError:
                        status = None
                        continue

                    value, exists = local.read(key, context)
                    assert exists, exists
                    return value, True

        return None, status

    def fetch(self, keys: Iterable[Key], context, *, verbose: bool) -> Sequence[Key]:
        def store(k, base):
            status = self._replicate(k, base, context)
            bar.update()
            return status if status is WriteError else k

        keys = set(keys)
        bar = tqdm(disable=not verbose, total=len(keys))
        present = set()
        for local in self.local:
            for key in keys:
                if local.contains(key, context):
                    present.add(key)
                    bar.update()

        keys -= present
        logger.info(f'Fetch: {len(present)} keys already present, fetching {len(keys)}')

        for remote in self.remote:
            if not keys:
                break

            logger.info(f'Trying remote {remote}')
            keys -= {
                k for k, success in remote.fetch(list(keys), store, self.hash)
                if success and k is not WriteError
            }

        return list(keys)

    def _replicate(self, key, base, context):
        for local in self.local:
            if local.replicate(key, base, context):
                return local

        return WriteError



import os
import shutil
from enum import Enum
from pathlib import Path
from typing import Union


class Reason(Enum):
    WrongDigestSize, WrongHash, CorruptedHash, WrongFolderStructure, CorruptedData, Expired, Filtered = range(7)


def touch(path):
    os.utime(path)


def to_read_only(path: Path, permissions, group):
    os.chmod(path, 0o444 & permissions)
    shutil.chown(path, group=group)


def get_size(file: Path) -> int:
    return file.stat().st_size


def mkdir(path: Path, permissions: Union[int, None], group: Union[str, int, None],
          parents: bool = False, exist_ok: bool = False):
    path.mkdir(parents=parents, exist_ok=exist_ok)
    if permissions is not None:
        path.chmod(permissions)
    if group is not None:
        shutil.chown(path, group=group)


def create_folders(path: Path, permissions, group):
    if not path.exists():
        create_folders(path.parent, permissions, group)
        mkdir(path, permissions, group)



from typing import Callable, Sequence, Any

from tqdm import tqdm

from .base import EdgesBag, Container
from .cache import IdentityContext
from ..engine.base import Node, TreeNode
from ..engine.edges import FunctionEdge, StaticEdge, StaticGraph
from ..engine.graph import Graph
from ..engine.node_hash import FilterHash
from ..utils import extract_signature, node_to_dict


class FilterContainer(Container):
    """
    Changes only the `keys` attribute.
    """

    # TODO: remove default
    def __init__(self, predicate: Callable, verbose: bool, keys: str = 'ids'):
        self.names, _ = extract_signature(predicate)
        assert keys not in self.names
        self.predicate = predicate
        self.verbose = verbose
        self.keys = keys

    @staticmethod
    def _find(nodes, name):
        for node in nodes:
            if node.name == name:
                return node

        raise ValueError(f'The previous layer must contain the attribute "{name}"')

    def _make_graph(self, layer):
        copy = layer.freeze()
        edges = list(copy.edges)
        outputs_mapping = node_to_dict(copy.outputs)
        out = Node('$predicate')
        edges.append(
            FunctionEdge(self.predicate, len(self.names)).bind([outputs_mapping[name] for name in self.names], out))
        mapping = TreeNode.from_edges(edges)
        return Graph([mapping[copy.inputs[0]]], mapping[out])

    def wrap(self, container: EdgesBag) -> EdgesBag:
        main = container.freeze()
        outputs = list(main.outputs)
        edges = list(main.edges)

        # change ids
        keys = self._find(outputs, self.keys)
        out = Node(self.keys)
        outputs.remove(keys)
        outputs.append(out)

        # filter
        graph = self._make_graph(container)
        edges.append(FilterEdge(graph, self.verbose).bind(keys, out))
        return EdgesBag(main.inputs, outputs, edges, IdentityContext(), persistent_nodes=main.persistent_nodes)


class FilterEdge(StaticGraph, StaticEdge):
    def __init__(self, graph: Graph, verbose: bool):
        super().__init__(arity=1)
        self.verbose = verbose
        self.graph = graph
        self._hash = self.graph.hash()

    def _make_hash(self, hashes):
        keys, = hashes
        return FilterHash(self._hash, keys)

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        keys, = inputs
        return tuple(filter(self.graph.call, tqdm(
            keys, desc='Filtering', disable=not self.verbose,
        )))



import logging
import warnings
from pathlib import Path
from typing import Sequence, Iterable, Callable, Any

from .digest import digest_file
from .disk import Disk
from .interface import HashStorage, Key, RemoteStorage, ReadError, WriteError
from ..utils import PathLike

logger = logging.getLogger(__name__)


class Storage:
    def __init__(self, local: Sequence[Disk], remote: Sequence[RemoteStorage] = ()):
        self.storage = HashStorage(local, remote)
        self.algorithm = self.storage.hash.build()
        self.digest_size = sum(self.storage.local[0].levels)

    @property
    def local(self):
        return self.storage.local

    def read(self, func: Callable, key: Key, *args, fetch: bool = True, **kwargs):
        value, success = self.storage.read(key, lambda x: func(x, *args, **kwargs), fetch=fetch)
        if success:
            return value

        if success is None:
            raise WriteError(f"The key {key} couldn't be written to any storage")
        if fetch:
            message = f'Key {key} is not present neither locally nor among your {len(self.storage.remote)} remotes'
        else:
            message = f'Key {key} is not present locally'
        raise ReadError(message)

    def write(self, file: PathLike) -> Key:
        file = Path(file)
        assert file.exists(), file
        key = digest_file(file, self.algorithm)
        if not self.storage.write(key, file, None):
            raise WriteError('The file could not be written to any storage')

        return key

    def fetch(self, keys: Iterable[Key], *, verbose: bool) -> Sequence[Key]:
        return self.storage.fetch(keys, None, verbose=verbose)

    def resolve(self, key: Key, *, fetch: bool = True) -> Path:
        """ This is not safe, but it's fast. """
        return self.read(lambda path: path, key, fetch=fetch)

    # deprecated names
    def get_path(self, key: Key, fetch: bool = True) -> Path:
        warnings.warn('The method `get_path` was renamed to `resolve`', UserWarning, 2)
        warnings.warn('The method `get_path` was renamed to `resolve`', DeprecationWarning, 2)
        return self.resolve(key, fetch=fetch)

    def load(self, func: Callable, key: Key, *args, fetch: bool = True, **kwargs) -> Any:
        warnings.warn('The method `load` was renamed to `read`', UserWarning, 2)
        warnings.warn('The method `load` was renamed to `read`', DeprecationWarning, 2)
        return self.read(func, key, *args, fetch=fetch, **kwargs)

    def store(self, file: PathLike) -> Key:
        warnings.warn('The method `store` was renamed to `write`', UserWarning, 2)
        warnings.warn('The method `store` was renamed to `write`', DeprecationWarning, 2)
        return self.write(file)



import logging
import time
from abc import ABC, abstractmethod
from contextlib import contextmanager
from threading import Lock
from typing import ContextManager, MutableMapping

from redis import Redis

Key = str
logger = logging.getLogger(__name__)


class Locker(ABC):
    def __init__(self, track_size: bool):
        self.track_size = track_size

    @contextmanager
    def read(self, key: Key):
        self.reserve_read(key)
        try:
            yield
        finally:
            self.stop_reading(key)

    @contextmanager
    def write(self, key: Key):
        self.reserve_write(key)
        try:
            yield
        finally:
            self.stop_writing(key)

    def reserve_read(self, key: Key):
        sleep_time = 0.1
        sleep_iters = int(600 / sleep_time) or 1  # 10 minutes
        wait_for_true(self.start_reading, key, sleep_time, sleep_iters)

    def reserve_write(self, key: Key):
        sleep_time = 0.1
        sleep_iters = int(600 / sleep_time) or 1  # 10 minutes
        wait_for_true(self.start_writing, key, sleep_time, sleep_iters)

    @abstractmethod
    def start_reading(self, key: Key) -> bool:
        """ Try to reserve a read operation. Return True if it was successful. """

    @abstractmethod
    def stop_reading(self, key: Key):
        """ Release a read operation. """

    @abstractmethod
    def start_writing(self, key: Key) -> bool:
        """ Try to reserve a write operation. Return True if it was successful. """

    @abstractmethod
    def stop_writing(self, key: Key):
        """ Release a write operation. """

    # TODO: move this to another interface?
    def get_size(self):
        raise NotImplementedError

    def inc_size(self, size: int):
        raise NotImplementedError

    def dec_size(self, size: int):
        raise NotImplementedError

    def set_size(self, size: int):
        raise NotImplementedError


class DummyLocker(Locker):
    def __init__(self):
        super().__init__(False)

    def start_reading(self, key: Key) -> bool:
        return True

    def stop_reading(self, key: Key):
        pass

    def start_writing(self, key: Key) -> bool:
        return True

    def stop_writing(self, key: Key):
        pass


class DictRegistry:
    _lock: ContextManager
    _reading: MutableMapping[Key, int]
    _writing: MutableMapping[Key, int]

    def _get_reading(self, key):
        value = self._reading.get(key, 0)
        logger.info(f'Read count {value}')
        assert value >= 0, value
        return value

    def _get_writing(self, key):
        value = self._writing.get(key, 0)
        logger.info(f'Write count {value}')
        assert 0 <= value <= 1, value
        return value

    def _is_reading(self, key: Key):
        return bool(self._get_reading(key))

    def _is_writing(self, key: Key):
        return bool(self._get_writing(key))

    def start_reading(self, key: Key) -> bool:
        with self._lock:
            if self._is_writing(key):
                return False

            self._reading[key] = self._get_reading(key) + 1
            return True

    def stop_reading(self, key: Key):
        with self._lock:
            value = self._get_reading(key)
            if value == 1:
                self._reading.pop(key)
            else:
                self._reading[key] = value - 1

    def start_writing(self, key: Key) -> bool:
        with self._lock:
            if self._is_reading(key) or self._is_writing(key):
                return False

            value = self._get_writing(key)
            assert value == 0, value
            self._writing[key] = value + 1
            return True

    def stop_writing(self, key: Key):
        with self._lock:
            value = self._get_writing(key)
            assert value == 1, value
            self._writing.pop(key)


class ThreadLocker(DictRegistry, Locker):
    def __init__(self):
        super().__init__(False)
        self._lock = Lock()
        self._reading = {}
        self._writing = {}


class GlobalThreadLocker(Locker):
    def __init__(self):
        super().__init__(False)
        self._lock = Lock()

    def _acquire(self):
        if self._lock.locked():
            return False
        self._lock.acquire()
        return True

    def _release(self):
        assert self._lock.locked()
        self._lock.release()

    def start_reading(self, key: Key) -> bool:
        return self._acquire()

    def stop_reading(self, key: Key):
        self._release()

    def start_writing(self, key: Key) -> bool:
        return self._acquire()

    def stop_writing(self, key: Key):
        self._release()


class RedisLocker(Locker):
    def __init__(self, *args, prefix: str, expire: int):
        super().__init__(True)
        if len(args) == 1 and isinstance(args[0], Redis):
            redis, = args
        else:
            redis = Redis(*args)

        self._redis = redis
        self._prefix = prefix + ':'
        self._expire = expire
        self._volume_key = f'{prefix}.V'
        # TODO: how slow are these checks?
        # language=Lua
        self._stop_writing = self._redis.script_load('''
        if redis.call('get', KEYS[1]) == '-1' then
            redis.call('del', KEYS[1])
        else
            error('')
        end''')
        # language=Lua
        self._start_reading = self._redis.script_load(f'''
        local lock = redis.call('get', KEYS[1])
        if lock == '-1' then 
            return 0
        elseif lock == false then
            redis.call('set', KEYS[1], 1, 'EX', {expire})
            return 1
        else
            redis.call('set', KEYS[1], lock + 1, 'EX', {expire})
            return 1
        end''')
        # language=Lua
        self._stop_reading = self._redis.script_load(f'''
        local lock = redis.call('get', KEYS[1])
        if lock == '1' then
            redis.call('del', KEYS[1])
        elseif tonumber(lock) < 1 then
            error('')
        else
            redis.call('set', KEYS[1], lock - 1, 'EX', {expire})
        end''')

    def start_writing(self, key: Key) -> bool:
        return bool(self._redis.set(self._prefix + key, -1, nx=True, ex=self._expire))

    def stop_writing(self, key: Key):
        self._redis.evalsha(self._stop_writing, 1, self._prefix + key)

    def start_reading(self, key: Key) -> bool:
        return bool(self._redis.evalsha(self._start_reading, 1, self._prefix + key))

    def stop_reading(self, key: Key):
        self._redis.evalsha(self._stop_reading, 1, self._prefix + key)

    def get_size(self):
        return int(self._redis.get(self._volume_key) or 0)

    def set_size(self, size: int):
        self._redis.set(self._volume_key, size)

    def inc_size(self, size: int):
        self._redis.incrby(self._volume_key, size)

    def dec_size(self, size: int):
        self._redis.decrby(self._volume_key, size)

    @classmethod
    def from_url(cls, url: str, prefix: str, expire: int):
        return cls(Redis.from_url(url), prefix=prefix, expire=expire)


class PotentialDeadLock(RuntimeError):
    pass


def wait_for_true(func, key, sleep_time, max_iterations):
    i = 0
    while not func(key):
        if i >= max_iterations:
            logger.error('Potential deadlock detected for %s', key)
            raise PotentialDeadLock(f"It seems like you've hit a deadlock for key {key}.")

        time.sleep(sleep_time)
        i += 1

    logger.debug('Waited for %d iterations for %s', i, key)



import filecmp
import logging
import errno
import shutil
from datetime import datetime
from pathlib import Path
from typing import Set, Union, Any, Tuple, Callable

from .digest import digest_file, get_digest_size
from .interface import Key
from .local import DiskBase
from .utils import to_read_only, Reason

FILENAME = 'data'
logger = logging.getLogger(__name__)


class Disk(DiskBase):
    def read(self, key, context: Callable) -> Tuple[Any, bool]:
        base = self._key_to_base(key)

        with self.locker.read(key):
            if not base.exists():
                return None, False

            return context(base / FILENAME), True

    def _check_consistency(self, base: Path, key: Key, value: Path, context):
        match_files(value, self._key_to_base(key) / FILENAME)

    def _write(self, base: Path, key: Key, value: Path, context):
        value = Path(value)
        assert value.is_file(), value

        file = base / FILENAME
        copy_file(value, file)
        # make file read-only
        to_read_only(file, self.permissions, self.group)
        digest = digest_file(file, self.algorithm)
        if digest != key:
            shutil.rmtree(base)
            raise ValueError(f'The stored file has a wrong hash: expected {key} got {digest}. '
                             'The file was most likely corrupted while copying')

    def _replicate(self, base: Path, key: Key, source: Path, context):
        self._write(base, key, source / FILENAME, context)

    def inspect_entry(self, key: Key, allowed_keys: Set[Key] = None, created: Union[float, datetime] = None):
        if len(key) != get_digest_size(self.levels, string=True):
            return Reason.WrongDigestSize

        base = self._key_to_base(key)

        # we remove missing hashes only if they are older than a given age
        if allowed_keys and key not in allowed_keys:
            if created is None:
                return Reason.Filtered
            if isinstance(created, datetime):
                created = created.timestamp()
            if base.stat().st_mtime < created:
                return Reason.Filtered

        with self.locker.read(key):
            if not (base / FILENAME).exists():
                return Reason.CorruptedData

            if digest_file(base / FILENAME, self.algorithm) != key:
                return Reason.WrongHash


def copy_file(source, destination):
    # in Python>=3.8 the sendfile call is used, which apparently may fail
    try:
        shutil.copyfile(source, destination)
    except OSError as e:
        # BlockingIOError -> fallback to slow copy
        if e.errno != errno.EWOULDBLOCK:
            raise

        with open(source, 'rb') as src, open(destination, 'wb') as dst:
            shutil.copyfileobj(src, dst)


def match_files(first: Path, second: Path):
    if not filecmp.cmp(first, second, shallow=False):
        raise ValueError(f'Files do not match: {first} vs {second}')



from abc import ABC, abstractmethod
from itertools import starmap
from math import ceil
from typing import Any, Generator, Union

from tqdm import tqdm

from .base import Nodes, Tuple, BoundEdges, EdgesBag, Container, Context
from ..engine import NodeHash
from ..engine.base import Node, TreeNode, Edge, HashOutput, Request, Response, Command
from ..engine.edges import CacheEdge, IdentityEdge, ImpureEdge
from ..engine.graph import Graph
from ..engine.node_hash import NodeHashes, TupleHash
from ..exceptions import DependencyError
from ..utils import node_to_dict
from ..cache import Cache, MemoryCache, DiskCache


class IdentityContext(Context):
    def reverse(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges) -> Tuple[Nodes, BoundEdges]:
        # just propagate everything
        return outputs, edges

    def update(self, mapping: dict) -> 'Context':
        return self


class CacheBase(Container):
    pass


class CacheContainer(CacheBase, ABC):
    def __init__(self, names, allow_impure):
        self.cache_names = names
        self.allow_impure = allow_impure

    @abstractmethod
    def get_storage(self) -> Cache:
        """ Create a cache storage instance """

    def wrap(self, container: 'EdgesBag') -> 'EdgesBag':
        state = container.freeze()
        forward_outputs = node_to_dict(state.outputs)

        edges = list(state.edges)
        outputs = [Node(name) for name in forward_outputs]
        mapping = TreeNode.from_edges(state.edges)

        for node in outputs:
            if self.cache_names is None or node.name in self.cache_names:
                if not self.allow_impure:
                    self._detect_impure(mapping[forward_outputs[node.name]], node.name)
                edges.append(CacheEdge(self.get_storage()).bind(forward_outputs[node.name], node))
            else:
                edges.append(IdentityEdge().bind(forward_outputs[node.name], node))

        return EdgesBag(state.inputs, outputs, edges, IdentityContext(), persistent_nodes=state.persistent_nodes)

    @staticmethod
    def _detect_impure(node: TreeNode, name: str):
        if node.is_leaf:
            return

        if isinstance(node.edge, ImpureEdge):
            raise ValueError(f'You are trying to cache the field "{name}", '
                             f'which has an `impure` dependency - "{node.name}"')

        for parent in node.parents:
            CacheContainer._detect_impure(parent, name)


class MemoryCacheContainer(CacheContainer):
    def __init__(self, names, size, allow_impure, cache_instances):
        super().__init__(names, allow_impure)
        self.cache_instances = cache_instances
        self.size = size

    def get_storage(self):
        cache = MemoryCache(self.size)
        self.cache_instances.add(cache)
        return cache


class DiskCacheContainer(CacheContainer):
    def __init__(self, names, local, remote, allow_impure):
        super().__init__(names, allow_impure)
        self.storage = DiskCache(local, remote, fetch=bool(remote))

    def get_storage(self):
        return self.storage


class CacheColumnsContainer(CacheBase):
    """
    CacheRow = Product + CacheToDisk + CacheToRam + Projection
    """

    def __init__(self, names, local, remote, verbose, shard_size):
        self.shard_size = shard_size
        self.verbose = verbose
        self.cache_names = names
        self.disk = DiskCache(local, remote, bool(remote))
        self.ram = MemoryCache(None)

    def wrap(self, container: EdgesBag) -> EdgesBag:
        main = container.freeze()
        edges = list(main.edges)
        key, = main.inputs
        main_outputs = node_to_dict(main.outputs)
        # TODO: containers must know about property names
        property_name = 'ids'
        if property_name not in main_outputs:
            raise DependencyError(f'The previous layer must contain the "{property_name}" property.')

        keys = main_outputs[property_name]

        copy = container.freeze()
        mapping = TreeNode.from_edges(copy.edges)
        outputs_copy = node_to_dict(copy.outputs)
        graph_inputs = [mapping[copy.inputs[0]]]

        outputs = []
        for output in main.outputs:
            name = output.name
            if name not in self.cache_names:
                outputs.append(output)

            else:
                local = Node(name)
                # build a graph for each node
                graph = Graph(graph_inputs, mapping[outputs_copy[name]])
                edges.append(CachedColumn(
                    self.disk, self.ram, graph, self.verbose, self.shard_size).bind([output, key, keys], local))
                outputs.append(local)

        return EdgesBag([key], outputs, edges, IdentityContext(), persistent_nodes=main.persistent_nodes)


class CachedColumn(Edge):
    """
    Edge Inputs
    -----------
    entry: the entry at ``key``
    key: a unique key for each entry in the tuple
    keys: all available keys
    """

    def __init__(self, disk: DiskCache, ram: MemoryCache, graph: Graph, verbose: bool,
                 shard_size: Union[int, float, None]):
        super().__init__(arity=3)
        self.graph = graph
        self.disk = disk
        self.ram = ram
        self.verbose = verbose
        self.shard_size = shard_size

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        # propagate the first value
        value = yield Command.ParentHash, 0
        return value, None

    def _get_shard(self, key, keys):
        keys = sorted(keys)
        assert key in keys

        size = self.shard_size
        if size is None:
            return keys, 1, 0
        if isinstance(size, float):
            size = ceil(size * len(keys))

        assert size > 0
        idx = keys.index(key) // size
        count = ceil(len(keys) / size)
        start = idx * size
        keys = keys[start:start + size]
        assert key in keys
        return keys, count, idx

    def evaluate(self) -> Generator[Request, Response, Any]:
        output = yield Command.CurrentHash,
        value, exists = self.ram.raw_get(output)
        if exists:
            return value

        key = yield Command.ParentValue, 1
        keys = yield Command.ParentValue, 2
        keys, shards_count, shard_idx = self._get_shard(key, keys)

        hashes, states = [], []
        for k in keys:
            h, state = self.graph.get_hash(k)
            hashes.append(h)
            states.append(state)
            if k == key:
                assert output == h, (output, h)
        # TODO: hash the graph?
        compound = TupleHash(*hashes)

        digest, context = self.disk.prepare(compound)
        values, exists = self.disk.get(digest, context)
        if not exists:
            suffix = '' if shards_count == 1 else f' ({shard_idx}/{shards_count})'
            values = tuple(starmap(self.graph.get_value, tqdm(
                states, desc=f'Generating the columns cache{suffix}', disable=not self.verbose,
            )))
            self.disk.set(digest, values, context)

        for k, h, value in zip(keys, hashes, values):
            self.ram.raw_set(h, value)
            if k == key:
                result = value

        return result  # noqa

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]



from typing import Union, Any, Tuple

from pylru import lrucache

from .base import Cache
from ..engine import NodeHash
from ..storage.locker import ThreadLocker, GlobalThreadLocker


class MemoryCache(Cache):
    def __init__(self, size: Union[int, None]):
        super().__init__()
        # TODO: maybe just always use a global lock without the Locker interface
        if size is not None:
            self._cache = lrucache(size)
            self.locker = GlobalThreadLocker()

        else:
            self._cache = {}
            self.locker = ThreadLocker()

    def get(self, key: NodeHash, context) -> Tuple[Any, bool]:
        key = key.value
        with self.locker.read(key):
            if key in self._cache:
                return self._cache[key], True
            return None, False

    def set(self, key: NodeHash, value: Any, context):
        key = key.value
        with self.locker.write(key):
            self._cache[key] = value

    def clear(self):
        # TODO: need a global lock
        self._cache = {}



"""
This set of functions was extracted from the `cloudpickle` package, for the sole purpose of
reproducibility across various versions of `cloudpickle`.

`cloudpickle` in turn is based on the `cloud` package, developed by `PiCloud, Inc.
<https://web.archive.org/web/20140626004012/http://www.picloud.com/>`_.

Copyright (c) 2012, Regents of the University of California.
Copyright (c) 2009 `PiCloud, Inc. <https://web.archive.org/web/20140626004012/http://www.picloud.com/>`_.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of California, Berkeley nor the
      names of its contributors may be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import importlib
import sys
import warnings
from weakref import WeakSet
from collections import ChainMap
from enum import Enum
from pickle import _Pickler as Pickler
from types import ModuleType
from typing import Union, Set

from cloudpickle import CloudPickler
from cloudpickle.cloudpickle import _whichmodule, _extract_code_globals, _get_cell_contents, _lookup_module_and_qualname

try:
    from gzip import BadGzipFile
except ImportError:
    BadGzipFile = OSError


class PickleMode(Enum):
    Global, Deep = range(2)


def get_pickle_mode(obj, name=None):
    if obj in STABLE_OBJECTS:
        return PickleMode.Global
    if obj in UNSTABLE_OBJECTS:
        return PickleMode.Deep

    pair = _lookup_module_and_qualname(obj, name)
    if pair is None:
        return PickleMode.Deep
    module, _ = pair
    module = module.__name__

    while True:
        if module in UNSTABLE_MODULES:
            return PickleMode.Deep

        split = module.rsplit(".", 1)
        if len(split) == 1:
            break
        module, _ = split

    _check_is_under_development(obj, name)
    return PickleMode.Global


def _check_is_under_development(obj, name):
    if name is None:
        name = getattr(obj, '__qualname__', None)
    if name is None:
        name = getattr(obj, '__name__', None)

    base_module = _whichmodule(obj, name).split('.', 1)[0]
    base = sys.modules.get(base_module)
    if base is None:
        base = importlib.import_module(base_module)

    if hasattr(base, '__development__'):
        raise RuntimeError(f'Error in "{base_module}": you are relying on an old cache invalidation machinery: '
                           'use `unstable_module(__name__)` instead')


def extract_func_data(func):
    """A simplified version of the same function from cloudpickle"""
    # `base globals` are not used - they are only needed for unpickling
    code = func.__code__
    func_global_refs = _extract_code_globals(code)
    f_globals = {}
    for var in func_global_refs:
        if var in func.__globals__:
            f_globals[var] = func.__globals__[var]

    defaults = func.__defaults__

    closure = (
        list(map(_get_cell_contents, func.__closure__))
        if func.__closure__ is not None
        else None
    )
    # save the dict
    dct = func.__dict__
    return code, f_globals, defaults, closure, dct


DISPATCH, DISPATCH_TABLE = Pickler.dispatch.copy(), {}
# various versions of cloudpickle store the reducers/dispatch table in different places
#  1. the simplest case:
if hasattr(CloudPickler, 'dispatch_table') and isinstance(CloudPickler.dispatch_table, (dict, ChainMap)):
    DISPATCH_TABLE = CloudPickler.dispatch_table.copy()
# 2. the `dispatch` dict can actually be a `dispatch_table`:
elif 'reduce' in CloudPickler.dispatch[classmethod].__name__:
    DISPATCH_TABLE = CloudPickler.dispatch.copy()
# 3. the `dispatch` is really a dispatch
else:
    DISPATCH.update(CloudPickler.dispatch)


def to_dispatch(func):
    def save(self, obj):
        args = func(obj)
        if isinstance(args, str):
            self.save_global(obj, args)
        else:
            self.save_reduce(*args, obj=obj)

    return save


DISPATCH.update({k: to_dispatch(v) for k, v in DISPATCH_TABLE.items()})
# we use a set of weak refs, because we don't want to cause memory leaks
STABLE_OBJECTS = WeakSet()
UNSTABLE_OBJECTS = WeakSet()
UNSTABLE_MODULES: Set[str] = set()


def is_stable(obj):
    """
    Decorator that opts out a function or class from being pickled during node hash calculation.
    Use it if you are sure that your function/class will never change in a way that might affect its behaviour.
    """
    if obj in UNSTABLE_OBJECTS:
        warnings.warn('The object was already marked as unstable')
        UNSTABLE_OBJECTS.remove(obj)
    STABLE_OBJECTS.add(obj)
    return obj


def is_unstable(obj):
    if obj in STABLE_OBJECTS:
        warnings.warn('The object was already marked as stable')
        STABLE_OBJECTS.remove(obj)
    UNSTABLE_OBJECTS.add(obj)
    return obj


def unstable_module(module: Union[str, ModuleType]):
    if not isinstance(module, str):
        module = module.__name__
    UNSTABLE_MODULES.add(module)



import logging
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from tqdm import tqdm

from .config import root_params, load_config, make_locker
from .digest import digest_to_relative
from .interface import LocalStorage, Key
from .utils import get_size, create_folders
from ..utils import PathLike

logger = logging.getLogger(__name__)


class DiskBase(LocalStorage, ABC):
    def __init__(self, root: PathLike):
        config = load_config(root)
        super().__init__(config.hash, config.levels)
        self.root = Path(root)
        self.permissions, self.group = root_params(self.root)

        self.locker = make_locker(config.locker)
        self.min_free_size = config.free_disk_size
        self.max_size = config.max_size

        if not self.locker.track_size:
            assert self.max_size is None or self.max_size == float('inf'), self.max_size

    def _key_to_base(self, key: Key):
        return self.root / digest_to_relative(key, self.levels)

    def _writeable(self):
        result = True

        if self.min_free_size > 0:
            result = result and shutil.disk_usage(self.root).free >= self.min_free_size

        if self.max_size is not None and self.max_size < float('inf'):
            result = result and self.locker.get_size() <= self.max_size

        return result

    @staticmethod
    def _get_size(base: Path):
        return sum(get_size(file) for file in base.glob('**/*') if file.is_file())

    def _protected_write(self, key, value: Any, context, write) -> bool:
        base = self._key_to_base(key)
        with self.locker.write(key):
            # if already stored
            if base.exists():
                self._check_consistency(base, key, value, context)
                return True

            # make sure we can write
            if not self._writeable():
                return False

            try:
                # create base folder
                create_folders(base, self.permissions, self.group)
                # populate the folder
                write(base, key, value, context)
                # increase the storage size
                if self.locker.track_size:
                    self.locker.inc_size(self._get_size(base))

            except BaseException as e:
                if base.exists():
                    shutil.rmtree(base)
                raise RuntimeError('An error occurred while copying the file') from e

            return True

    def write(self, key, value: Any, context: Any) -> bool:
        return self._protected_write(key, value, context, self._write)

    def replicate(self, key, source: Path, context: Any) -> bool:
        return self._protected_write(key, source, context, self._replicate)

    def remove(self, key: Key) -> bool:
        base = self._key_to_base(key)
        with self.locker.write(key):
            if not base.exists():
                return False

            size = self._get_size(base)
            shutil.rmtree(base)
            if self.locker.track_size:
                self.locker.dec_size(size)

            return True

    def contains(self, key: Key, context):
        """ This is not safe, but it's fast. """
        with self.locker.read(key):
            return self._key_to_base(key).exists()

    def actualize(self, verbose: bool):
        """ Useful for migration between locking mechanisms. """
        # TODO: need a global lock
        size = 0
        bar = tqdm(self.root.glob(f'**/*'), disable=not verbose)
        for file in bar:
            if file.is_dir():
                continue

            bar.set_description(str(file.parent.relative_to(self.root)))
            assert file.is_file()
            size += get_size(file)

        self.locker.set_size(size)

    @abstractmethod
    def _check_consistency(self, base: Path, key: Key, value: Any, context):
        pass

    @abstractmethod
    def _write(self, base: Path, key: Key, value: Any, context):
        pass

    @abstractmethod
    def _replicate(self, base: Path, key: Key, source: Path, context):
        pass



import logging
from abc import ABC, abstractmethod
from operator import itemgetter
from typing import Tuple, Optional, Set, AbstractSet

from ..engine.edges import FunctionEdge, ProductEdge, IdentityEdge
from ..engine.graph import Graph
from ..engine.base import TreeNode, BoundEdge, Node, Nodes, BoundEdges
from ..engine import Backend, DefaultBackend
from ..exceptions import GraphError, FieldError
from ..utils import node_to_dict

logger = logging.getLogger(__name__)

NameSet = AbstractSet[str]


class Container:
    def wrap(self, container: 'EdgesBag') -> 'EdgesBag':
        raise NotImplementedError


class GraphCompiler:
    def __init__(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges, virtuals: NameSet, backend: Backend):
        tree_node_map = TreeNode.from_edges(edges)
        self._edges = edges
        self.inputs = tuple(tree_node_map[x] for x in inputs)
        self.outputs = node_to_dict(tree_node_map[x] for x in outputs)
        self.backend = backend
        self.virtuals = virtuals
        self.methods = {node.name: self._compile(node) for node in self.outputs.values()}

    def __getitem__(self, item):
        if item not in self.methods:
            if isinstance(item, str) and item in self.virtuals:
                # TODO: signature
                value = identity

            elif isinstance(item, tuple):
                inputs, outputs = [], []
                for name in item:
                    if name not in self.outputs:
                        if name in self.virtuals:
                            output = TreeNode(name, None)
                            inputs.append(output)
                        else:
                            raise FieldError(f'"{name}" is not an available output: {tuple(self.outputs)}')
                    else:
                        output = self.outputs[name]
                    outputs.append(output)

                product = TreeNode('$product', (ProductEdge(len(item)), outputs))
                value = self._compile(product, tuple(inputs))

            else:
                raise FieldError(f'"{item}" is not an available output: {tuple(self.outputs)}')

            self.methods[item] = value

        return self.methods[item]

    def _compile(self, node, inputs=()):
        return Graph(self.inputs + inputs, node, self.backend).call


class Context(ABC):
    @abstractmethod
    def reverse(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges) -> Tuple[Nodes, BoundEdges]:
        pass

    @abstractmethod
    def update(self, mapping: dict) -> 'Context':
        pass


class NoContext(Context):
    def reverse(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges) -> Tuple[Nodes, BoundEdges]:
        raise ValueError('The layer is not reversible')

    def update(self, mapping: dict) -> 'Context':
        return self


class EdgesBag(Container):
    def __init__(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges, context: Optional[Context], *,
                 virtual_nodes: NameSet = None, persistent_nodes: Optional[NameSet]):
        if virtual_nodes is None:
            virtual_nodes = set()
        if persistent_nodes is None:
            persistent_nodes = set()
        if context is None:
            context = NoContext()

        self.inputs, self.outputs, self.edges, self.virtual_nodes = normalize_bag(
            inputs, outputs, edges, virtual_nodes, persistent_nodes)

        self.persistent_nodes = persistent_nodes
        self.context = context
        self.backend = DefaultBackend

    def freeze(self) -> 'EdgesBag':
        # TODO: layer inputs and outputs may not be among the edges
        node_map = {}
        edges_copy = []

        for edge in self.edges:
            inputs = update_map(edge.inputs, node_map)
            output, = update_map([edge.output], node_map)
            edges_copy.append(BoundEdge(edge.edge, inputs, output))

        return EdgesBag(
            update_map(self.inputs, node_map),
            update_map(self.outputs, node_map),
            edges_copy,
            self.context.update(node_map),
            virtual_nodes=self.virtual_nodes, persistent_nodes=self.persistent_nodes
        )

    def compile(self) -> GraphCompiler:
        return GraphCompiler(self.inputs, self.outputs, self.edges, self.virtual_nodes, self.backend)

    def loopback(self, func, inputs, output):
        state = self.freeze()
        edges = list(state.edges)
        current = node_to_dict(state.outputs)

        # connect forward outputs with bridge
        outputs = []
        if isinstance(inputs, str):
            inputs = inputs,
        inputs = tuple(inputs)

        if len(set(inputs)) != len(inputs):
            raise ValueError(f'The inputs contain duplicates: {inputs}')

        input_nodes, all_inputs = [], list(state.inputs)
        for name in inputs:
            if name not in current:
                if name not in state.virtual_nodes:
                    raise GraphError(f'Node "{name}" is not defined')
                node = Node(name)
                all_inputs.append(node)
            else:
                node = current[name]

            input_nodes.append(node)

        edge = FunctionEdge(func, len(input_nodes))

        # single output
        if isinstance(output, str):
            output = Node(output)

            edges.append(edge.bind(input_nodes, output))
            outputs.append(output)

        # multiple outputs
        else:
            assert len(set(outputs)) == len(outputs)

            aux = Node('$aux')
            edges.append(edge.bind(input_nodes, aux))
            for idx, out in enumerate(output):
                out = Node(out)
                edges.append(FunctionEdge(itemgetter(idx), 1).bind(aux, out))
                outputs.append(out)

        outputs, edges = state.context.reverse(all_inputs, outputs, edges)
        return GraphCompiler(all_inputs, outputs, edges, set(), self.backend)


def update_map(nodes, node_map):
    for node in nodes:
        if node not in node_map:
            node_map[node] = Node(node.name)
    return [node_map[x] for x in nodes]


def identity(x):
    return x


def get_parents(node: TreeNode):
    if node.is_leaf:
        return

    for parent in node.parents:
        if parent.is_leaf:
            yield parent
        else:
            yield from get_parents(parent)


def normalize_bag(inputs: Nodes, outputs: Nodes, edges: BoundEdges, virtual_nodes: Set[str], persistent_nodes):
    # 1. outputs must only depend on inputs
    # 1a. inputs must have no dependencies
    # 2. each node can only have a single incoming edge
    # 2a. the intersection between outputs and virtual nodes must be empty
    # 3. virtual edges with a present input node become non-virtual
    # 4. the graph must be acyclic
    inputs, outputs = node_to_dict(inputs), node_to_dict(outputs)
    edges = list(edges)

    # 2a:
    intersection = virtual_nodes & set(outputs)
    if intersection:
        raise GraphError(f'The nodes {intersection} are both inherited and have defined edges')

    # 3:
    add = (virtual_nodes | persistent_nodes) & set(inputs) - set(outputs)
    virtual_nodes = virtual_nodes - add
    for name in add:
        outputs[name] = Node(name)
        edges.append(IdentityEdge().bind(inputs[name], outputs[name]))

    # 2:
    adjacency = {}
    for edge in edges:
        if edge.output in adjacency:
            raise GraphError(f'The node "{edge.output.name}" has multiple incoming edges')
        adjacency[edge.output] = edge.inputs
    # 4:
    cycles = detect_cycles(adjacency)
    if cycles:
        raise GraphError(
            'The computational graph contains cycles:\n  ' +
            '\n  '.join(' -> '.join(node.name for node in nodes) for nodes in cycles)
        )

    # 1a:
    product = Node('$product')
    mapping = TreeNode.from_edges(edges + [ProductEdge(len(outputs)).bind(tuple(outputs.values()), product)])
    tree_inputs = {mapping[node] for node in inputs.values()}
    not_leaves = {node.output.name for node in tree_inputs if not node.is_leaf}
    if not_leaves:
        raise GraphError(f'The inputs {not_leaves} are not actual inputs - they have dependencies')
    # 1:
    missing = {node.name for node in set(get_parents(mapping[product])) - tree_inputs}
    if missing:
        raise GraphError(f'The nodes {missing} are missing from the inputs')

    return tuple(inputs.values()), tuple(outputs.values()), tuple(edges), virtual_nodes


def detect_cycles(adjacency):
    def visit(node):
        if node in visited:
            # the node was not completely visited
            if not visited[node]:
                cycles.append(tuple(stack) + (node,))

            return

        visited[node] = False
        stack.append(node)

        for parent in adjacency.get(node, ()):
            visit(parent)

        stack.pop()
        visited[node] = True

    stack, visited, cycles = [], {}, []
    for n in adjacency:
        visit(n)

    return cycles



from typing import Sequence, Callable, Tuple

NODE_TYPES = set()


class NodeHash:
    type: int

    def __init__(self, value):
        assert self.type is not None
        assert value[0] == self.type
        self.value = value

    def __eq__(self, other):
        return isinstance(other, NodeHash) and self.value == other.value

    def __init_subclass__(cls, **kwargs):
        if cls.type is not None:
            assert cls.type not in NODE_TYPES, (cls, cls.type, NODE_TYPES)
            NODE_TYPES.add(cls.type)


NodeHashes = Sequence[NodeHash]


class PrecomputeHash(NodeHash):
    type = None

    def __init__(self, value, hash_target):
        super().__init__(value)
        self._hash_target = hash_target
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self._hash_target)
        return self._hash


class CompoundBase(PrecomputeHash):
    type = None

    def __init__(self, *children: NodeHash):
        super().__init__(
            (self.type, *(h.value for h in children)),
            (self.type, *children),
        )


class LeafHash(PrecomputeHash):
    type = 0

    def __init__(self, data):
        value = self.type, data
        super().__init__(value, value)
        self.data = data


class ApplyHash(PrecomputeHash):
    type = 1

    def __init__(self, func: Callable, *args: NodeHash, kw_names: Tuple[str, ...]):
        # TODO: unify this during the next big migration
        values = tuple(h.value for h in args)
        kw_names = tuple(kw_names)
        if kw_names:
            seq = values, kw_names
        else:
            seq = values,

        super().__init__(
            (self.type, func, *seq),
            (self.type, func, args, kw_names),
        )


class GraphHash(CompoundBase):
    type = 2

    def __init__(self, output: NodeHash):
        super().__init__(output)


class TupleHash(CompoundBase):
    type = 3


# Higher order functions

class FilterHash(CompoundBase):
    type = 10

    def __init__(self, func: GraphHash, values: NodeHash):
        super().__init__(func, values)


class MergeHash(CompoundBase):
    type = 11


class GroupByHash(CompoundBase):
    type = 12

    def __init__(self, key: GraphHash, values: NodeHash):
        super().__init__(key, values)


class DictFromKeys(CompoundBase):
    type = 13

    def __init__(self, func: GraphHash, key: NodeHash, mapping: NodeHash):
        super().__init__(func, key, mapping)


class JoinMappingHash(PrecomputeHash):
    type = 14

    def __init__(self, left: GraphHash, right: GraphHash, left_keys: NodeHash, right_keys: NodeHash, id_maker):
        args = left, right, left_keys, right_keys
        super().__init__(
            (self.type, *tuple(h.value for h in args), id_maker),
            (self.type, *args, id_maker),
        )


# Experimental stuff

class MultiMappingHash(CompoundBase):
    type = -1



import logging
from typing import Callable, Iterable, TypeVar

from .compat import Generic
from .utils import format_arguments
from ..containers.base import Container, EdgesBag
from ..containers.pipeline import PipelineContainer, LazyPipelineContainer
from ..exceptions import FieldError
from ..utils import StringsLike

logger = logging.getLogger(__name__)

T = TypeVar('T', bound=Container)


class BaseLayer(Generic[T]):
    def __init__(self, container: T):
        self._container: T = container


class CallableLayer(BaseLayer[EdgesBag]):
    def __init__(self, container: EdgesBag, properties: Iterable[str]):
        super().__init__(container)
        self._methods = self._container.compile()
        self._properties = set(properties)

    def __getattr__(self, name):
        try:
            method = self._methods[name]
        except FieldError as e:
            raise AttributeError(name) from e

        if name in self._properties:
            return method()
        return method

    def __rshift__(self, layer: BaseLayer) -> 'Chain':
        return Chain(self, layer)

    def __call__(self, *args, **kwargs) -> 'Instance':
        return Instance(self, args, kwargs)

    def __dir__(self):
        return list(self._methods.outputs)

    def _wrap(self, func: Callable, inputs: StringsLike, outputs: StringsLike = None,
              final: StringsLike = None) -> Callable:
        return self._decorate(inputs, outputs, final)(func)

    def _decorate(self, inputs: StringsLike, outputs: StringsLike = None, final: StringsLike = None) -> Callable:
        if outputs is None:
            outputs = inputs
        if final is None:
            final = outputs
        if not isinstance(final, str):
            final = tuple(final)
        if isinstance(inputs, str):
            inputs = [inputs]

        def decorator(func: Callable) -> Callable:
            loopback = self._container.loopback(func, inputs, outputs)
            logger.info('Loopback compiled: %s', list(loopback.methods))
            return loopback[final]

        return decorator

    def _compile(self, inputs: StringsLike):
        if not isinstance(inputs, str):
            inputs = tuple(inputs)
        return self._methods[inputs]


class Instance:
    def __init__(self, layer: CallableLayer, args, kwargs):
        self._layer = layer
        self._args, self._kwargs = args, kwargs

    def _get(self, key):
        method = self._layer._compile(key)
        return method(*self._args, **self._kwargs)

    def __getattr__(self, name):
        return self._get(name)

    def __getitem__(self, name):
        return self._get(name)

    def __dir__(self):
        return dir(self._layer)


class Chain(CallableLayer):
    _container: PipelineContainer

    def __init__(self, head: CallableLayer, *tail: BaseLayer):
        super().__init__(
            PipelineContainer(head._container, *(layer._container for layer in tail)),
            head._properties,
        )
        self._layers = [head, *tail]

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._layers[index]

        if isinstance(index, slice):
            return Chain(*self._layers[index])

        raise ValueError('The index can be either an int or slice.')

    def _filter(self, func, *args, **kwargs):
        filtered = []

        for layer in self._layers:
            if isinstance(layer, Chain):
                filtered.append(layer._filter(func, *args, **kwargs))
            elif func(layer, *args, **kwargs):
                filtered.append(layer)

        return Chain(*filtered)

    def _filterfalse(self, func, *args, **kwargs):
        def not_func(x, *a, **kw):
            return not func(x, *a, **kw)

        return self._filter(not_func, *args, **kwargs)

    def _drop_cache(self):
        from .blocks import CacheLayer

        return self._filterfalse(isinstance, CacheLayer)

    def __repr__(self):
        if len(self._layers) == 2:
            a, b = self._layers
            return f'{a} >> {b}'

        return 'Chain' + format_arguments(self._layers)


class LazyChain(BaseLayer[LazyPipelineContainer]):
    def __init__(self, *layers: BaseLayer):
        super().__init__(LazyPipelineContainer(*(layer._container for layer in layers)))
        self._layers = layers

    def __repr__(self):
        return 'LazyChain' + format_arguments(self._layers)


def chained(*layers: BaseLayer, lazy: bool = False):
    base = LazyChain if lazy else Chain

    def decorator(klass):
        class Chained(base):
            def __init__(self, *args, **kwargs):
                super().__init__(klass(*args, **kwargs), *layers)

        return Chained

    return decorator



from collections import defaultdict
from typing import Callable, Generator, Any

from ..containers.base import EdgesBag
from ..engine import NodeHash
from ..engine.base import Node, Edge, Command, Request, Response, HashOutput, TreeNode
from ..engine.edges import StaticGraph, StaticHash, FunctionEdge, ProductEdge
from ..engine.graph import Graph
from ..engine.node_hash import LeafHash, NodeHashes, JoinMappingHash
from ..utils import Strings, node_to_dict


class JoinContainer(EdgesBag):
    def __init__(self, left: EdgesBag, right: EdgesBag, on: Strings, id_maker: Callable):
        assert len(set(on)) == len(on), on
        left, right = left.freeze(), right.freeze()

        edges = [*left.edges, *right.edges]
        outputs = []

        if len(left.inputs) != 1 or len(right.inputs) != 1:
            raise ValueError('Both layers should have exactly one input')

        # build the core mapping
        outputs_left, outputs_right = node_to_dict(left.outputs), node_to_dict(right.outputs)
        # TODO: parametrize these names
        keys_left, keys_right = outputs_left.pop('ids'), outputs_right.pop('ids')
        outputs_left.pop('id')
        outputs_right.pop('id')

        intersection = set(outputs_left) & set(outputs_right)
        missing = set(on) - intersection
        if missing:
            raise ValueError(f'Fields {missing} are missing')
        conflict = intersection - set(on)
        if conflict:
            raise ValueError(f'Field conflicts resolution not supported yet. Conflicts: {conflict}')

        mapping = Node('$mapping')
        edges.append(JoinMappingEdge(
            self._make_graph(left.inputs, outputs_left, left.edges, on),
            self._make_graph(right.inputs, outputs_right, right.edges, on),
            id_maker,
        ).bind([keys_left, keys_right], mapping))

        # add the new keys
        keys = Node('ids')
        outputs.append(keys)
        edges.append(FunctionEdge(lambda x: tuple(sorted(x)), 1).bind(mapping, keys))

        # add the new input
        inp = Node('id')
        edges.append(KeyProjection(0).bind([inp, mapping], left.inputs[0]))
        edges.append(KeyProjection(1).bind([inp, mapping], right.inputs[0]))

        # no_conflict = set(outputs_left) | set(outputs_right) - intersection
        outputs.extend(outputs_left.values())
        outputs.extend(outputs_right[x] for x in set(outputs_right) - set(outputs_left))
        super().__init__([inp], outputs, edges, None, persistent_nodes=left.persistent_nodes & right.persistent_nodes)

    @staticmethod
    def _make_graph(inputs, outputs, edges, on):
        edges = list(edges)
        output = Node('$output')
        edges.append(ProductEdge(len(on)).bind([outputs[x] for x in on], output))
        mapping = TreeNode.from_edges(edges)
        return Graph([mapping[x] for x in inputs], mapping[output])


class JoinMappingEdge(StaticGraph, StaticHash):
    def __init__(self, left: Graph, right: Graph, id_maker: Callable):
        super().__init__(arity=2)
        self.left = left
        self.right = right
        self.id_maker = id_maker
        self._hashes = left.hash(), right.hash()
        # TODO: this is potentially dangerous. should use a composition of Mapping and MemoryCache
        self._mapping = None

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return JoinMappingHash(*self._hashes, *inputs, id_maker=self.id_maker)

    def evaluate(self) -> Generator[Request, Response, Any]:
        if self._mapping is not None:
            return self._mapping

        left_keys, right_keys = yield Command.Await, (Command.ParentValue, 0), (Command.ParentValue, 1)
        precomputed = defaultdict(list)
        for i in left_keys:
            left = self.left.call(i)
            precomputed[left].append(i)

        mapping = {}
        for j in right_keys:
            right = self.right.call(j)
            for i in precomputed[right]:
                key = self.id_maker((i, j))
                if key in mapping:
                    raise ValueError(f'The provided key function is not reversible: value {key} already present')
                mapping[key] = i, j

        self._mapping = mapping
        return mapping


class KeyProjection(Edge):
    def __init__(self, index):
        super().__init__(2)
        self.index = index

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        key, mapping = yield Command.Await, (Command.ParentValue, 0), (Command.ParentValue, 1)
        value = mapping[key][self.index]
        return LeafHash(value), value

    def evaluate(self) -> Generator[Request, Response, Any]:
        value = yield Command.Payload,
        return value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        raise RuntimeError('Join cannot be currently a part of a subgraph')



import logging
from typing import Callable, Dict, Type, Union, Iterable

from .compat import SafeMeta
from ..utils import MultiDict
from .factory import SourceFactory, TransformFactory, FactoryLayer, add_from_mixins, add_quals, GraphFactory, \
    is_detectable

logger = logging.getLogger(__name__)

BASES: Dict[Type[FactoryLayer], GraphFactory] = {}


class APIMeta(SafeMeta):
    @classmethod
    def __prepare__(mcs, *args, **kwargs):
        return MultiDict()

    def __new__(mcs, class_name, bases, namespace, **flags):
        if '__factory' in flags:
            factory = flags.pop('__factory')
            assert bases == (FactoryLayer,)
            scope = namespace.to_dict()
            base = super().__new__(mcs, class_name, bases, scope, **flags)
            BASES[base] = factory
            return base

        bases = set(bases)
        intersection = set(BASES) & bases
        if len(intersection) != 1:
            raise TypeError(f'Layers must inherit from on of ' + ', '.join(x.__name__ for x in BASES))

        main, = intersection
        factory = BASES[main]
        bases -= intersection
        base_name = main.__name__
        for base in bases:
            if not issubclass(base, Mixin):
                raise TypeError(f'{base_name}s can only inherit directly from "{base_name}" or other mixins.')

        logger.info('Compiling the layer "%s" of type %s', class_name, base_name)

        if main == Mixin:
            add_from_mixins(namespace, bases)
            scope = add_quals({'__methods__': namespace}, namespace)

        else:
            factory.validate_before_mixins(namespace)
            add_from_mixins(namespace, bases)
            scope = factory.make_scope(namespace)

        return super().__new__(mcs, class_name, (main,), scope, **flags)


class Source(FactoryLayer, metaclass=APIMeta, __factory=SourceFactory):
    """
    Base class for all sources.
    """

    def __init__(self, *args, **kwargs):
        raise RuntimeError("\"Source\" can't be directly initialized. You must subclass it first.")


class Transform(FactoryLayer, metaclass=APIMeta, __factory=TransformFactory):
    """
    Base class for all transforms.

    Can also be used as an inplace factory for transforms.

    Examples
    --------
    # class-based transforms
    >>> class Zoom(Transform):
    >>>     def image(image):
    >>>         return zoom(image, scale_factor=2)
    # inplace transforms
    >>> Transform(image=lambda image: zoom(image, scale_factor=2))
    """
    __inherit__: Union[str, Iterable[str], bool] = ()

    def __init__(*args, __inherit__=(), **kwargs: Callable):
        assert args
        if len(args) > 1:
            raise TypeError('This constructor accepts only keyword arguments.')
        self, = args

        local = MultiDict()
        local['__inherit__'] = __inherit__
        for name, value in kwargs.items():
            if not is_detectable(value):
                raise TypeError(
                    f'All arguments for Transform must be callable. "{name}" is not callable but {type(value)}'
                )

            local[name] = value

        factory = TransformFactory(local)
        if factory.special_methods:
            raise TypeError(f"This constructor doesn't accept special methods: {tuple(factory.special_methods)}")
        super(Transform, self).__init__(factory.build({}), factory.property_names, ())

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._methods.methods)})"


class Mixin(FactoryLayer, metaclass=APIMeta, __factory=None):
    """
    Base class for all Mixins.
    """

    def __init__(*args, **kwargs):
        raise RuntimeError("Mixins can't be directly initialized.")

    __methods__: dict = {}



from pathlib import Path
from typing import Sequence


def digest_file(path: Path, algorithm, block_size=2 ** 20) -> str:
    hasher = algorithm()

    with open(path, 'rb') as f:
        while True:
            buffer = f.read(block_size)
            if not buffer:
                break
            hasher.update(buffer)

    return hasher.hexdigest()


def key_to_relative(key: str, levels: Sequence[int]):
    # TODO: too expensive?
    assert len(key) == get_digest_size(levels, string=True), len(key)

    parts = []
    start = 0
    for level in levels:
        stop = start + level * 2
        parts.append(key[start:stop])
        start = stop

    return Path(*parts)


digest_to_relative = key_to_relative


def get_digest_size(levels, string: bool):
    size = sum(levels)
    if string:
        size *= 2
    return size



from .base import Command
from .executor import Frame, Backend


# TODO: replace cache by a thunk tree
def execute(cmd, node, hashes, cache, backend: Backend):
    executor = backend.build(Frame([node], [(Command.Return,), (cmd,)], None))
    push, pop, peek = executor.push, executor.pop, executor.peek
    push_command, pop_command = executor.push_command, executor.pop_command
    next_frame, enqueue_frame = executor.next_frame, executor.enqueue_frame

    while True:
        if executor.frame.ready:
            next_frame()
            continue

        cmd, *args = pop_command()

        # return
        if cmd == Command.Return:
            assert not args
            assert len(executor.frame.stack) == 1, len(executor.frame.stack)
            executor.frame.value = pop()
            executor.frame.ready = True

            if executor.frame.parent is None:
                executor.clear()
                return executor.frame.value

            enqueue_frame(executor.frame.parent)
            next_frame()

        # communicate with edges
        elif cmd == Command.Send:
            node, iterator = args
            value = pop()
            try:
                request = iterator.send(value)

            except StopIteration as e:
                # clear the dependencies
                for n in node.parents:
                    hashes.evict(n)
                    cache.evict(n)
                # return value
                push(e.value)

            else:
                # must continue iteration
                push_command((cmd, node, iterator))
                push_command(request)
                push(node)

        # runs and caches `compute_hash`
        elif cmd == Command.ComputeHash:
            assert not args
            node = pop()
            if node in hashes:
                value = hashes[node]
                if isinstance(value, _CacheWaiter):
                    # restore state
                    push_command((cmd, *args))
                    push(node)
                    # remember to come back
                    value.add(executor.frame)
                    # switch context
                    next_frame()

                else:
                    push(value)
            else:
                hashes[node] = _CacheWaiter()
                push_command((Command.Store, hashes, node))
                push_command((Command.Send, node, node.edge.compute_hash()))
                push(None)

        # runs and caches `evaluate`
        elif cmd == Command.Evaluate:
            assert not args
            node = pop()

            if node in cache:
                value = cache[node]
                if isinstance(value, _CacheWaiter):
                    # restore state
                    push_command((cmd, *args))
                    push(node)
                    # remember to come back
                    value.add(executor.frame)
                    # switch context
                    next_frame()

                else:
                    push(value)
            else:
                cache[node] = _CacheWaiter()
                push_command((Command.Store, cache, node))
                push_command((Command.Send, node, node.edge.evaluate()))
                push(None)

        # requests
        elif cmd == Command.ParentHash:
            idx, = args
            node = pop()

            push_command((Command.Item, 0))
            push_command((Command.ComputeHash,))
            push(node.parents[idx])

        elif cmd == Command.ParentValue:
            idx, = args
            node = pop()

            push_command((Command.Evaluate,))
            push(node.parents[idx])

        elif cmd == Command.CurrentHash:
            assert not args
            push_command((Command.Item, 0))
            push_command((Command.ComputeHash,))

        elif cmd == Command.Payload:
            assert not args
            push_command((Command.Item, 1))
            push_command((Command.ComputeHash,))

        elif cmd == Command.Await:
            node = pop()
            push_command((Command.Tuple, len(args)))
            for arg in args:
                local = Frame([node], [(Command.Return,), arg], executor.frame)
                push_command((Command.AwaitThunk, local))
                enqueue_frame(local)

        elif cmd == Command.Call:
            pop()  # pop the node
            func, pos, kw = args
            executor.call(func, pos, kw)

        # utils
        elif cmd == Command.Store:
            storage, key = args
            parents = storage[key]
            assert isinstance(parents, _CacheWaiter), parents
            for parent in parents:
                enqueue_frame(parent)

            storage[key] = peek()

        elif cmd == Command.Item:
            key, = args
            push(pop()[key])

        elif cmd == Command.Tuple:
            n, = args
            push(tuple(pop() for _ in range(n)))

        elif cmd == Command.AwaitThunk:
            child, = args
            if child.ready:
                if child.error is not None:
                    raise child.error

                push(child.value)
            else:
                push_command((cmd, *args))
                next_frame()

        else:
            raise RuntimeError('Unknown command', cmd)


class _CacheWaiter(set):
    pass



from typing import Tuple, Sequence, NamedTuple, List, Dict, Set

from .base import Context, EdgesBag, update_map, NameSet
from ..engine.base import BoundEdge, Node, Nodes, BoundEdges, TreeNode, TreeNodes
from ..engine.edges import IdentityEdge
from ..engine.graph import count_entries
from ..exceptions import DependencyError
from ..utils import AntiSet, check_for_duplicates, node_to_dict


class LayerConnectionState(NamedTuple):
    # elements of composed layer
    inputs: List
    outputs: List
    edges: List[BoundEdge]
    # names of already used nodes
    used_output_names: set
    used_input_names: set
    cur_used_virtual: set
    prev_used_virtual: set
    # elements of current/previous layer
    cur_virtual: Set[str]
    cur_inputs: Dict[str, Node]
    cur_optional: Set[str]
    prev_virtual: set
    prev_persistent: Set[str]
    prev_outputs: Dict[str, Node]


class TransformContainer(EdgesBag):
    def __init__(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges, backward_inputs: Nodes = (),
                 backward_outputs: Nodes = (), *, optional_nodes: NameSet = None,
                 forward_virtual: NameSet, backward_virtual: NameSet, persistent_nodes: NameSet = None):

        forward_virtual, valid = normalize_inherit(forward_virtual, node_to_dict(outputs))
        assert valid
        backward_virtual, valid = normalize_inherit(backward_virtual, node_to_dict(backward_outputs))
        assert valid

        check_for_duplicates(inputs)
        super().__init__(
            inputs, outputs, edges,
            BagContext(backward_inputs, backward_outputs, backward_virtual),
            virtual_nodes=forward_virtual, persistent_nodes=persistent_nodes
        )
        self.optional_nodes = optional_nodes if optional_nodes is not None else set()

    def wrap(self, container: 'EdgesBag') -> 'EdgesBag':
        current = self.freeze()
        previous = container.freeze()
        all_edges = list(previous.edges) + list(current.edges)
        persistent_nodes = set.union(set(previous.persistent_nodes), set(self.persistent_nodes))
        state = LayerConnectionState(
            edges=all_edges,
            cur_used_virtual=set(),
            used_input_names=set(),
            prev_used_virtual=set(),
            used_output_names=set(),
            inputs=list(previous.inputs),
            outputs=list(current.outputs),
            cur_virtual=self.virtual_nodes,
            cur_optional=self.optional_nodes,
            cur_inputs=node_to_dict(current.inputs),
            prev_virtual=container.virtual_nodes,
            prev_outputs=node_to_dict(previous.outputs),
            prev_persistent=previous.persistent_nodes,
        )

        self._connect_common_nodes(state)
        self._connect_cur_virtual(state)
        self._connect_prev_virtual(state)

        essential_input_names = self.get_essential_input_names(current.inputs, current.outputs, current.edges)
        for o in current.outputs:
            # drop nodes that depend on inactive inputs
            if any(name not in state.used_input_names for name in essential_input_names[o]):
                state.outputs.remove(o)

        check_for_duplicates(state.outputs)
        new_virtual_nodes = self._merge_virtual_nodes(state)
        return EdgesBag(
            state.inputs, state.outputs, state.edges,
            PipelineContext(previous.context, current.context),
            virtual_nodes=new_virtual_nodes, persistent_nodes=persistent_nodes
        )

    @staticmethod
    def _connect_common_nodes(state: LayerConnectionState):
        cur_inputs = state.cur_inputs
        prev_outputs = state.prev_outputs
        common_node_names = set(cur_inputs.keys()).intersection(set(prev_outputs.keys()))
        for name in common_node_names:
            state.used_input_names.add(name)
            state.used_output_names.add(name)
            state.edges.append(IdentityEdge().bind(prev_outputs[name], cur_inputs[name]))

    @staticmethod
    def _connect_cur_virtual(state: LayerConnectionState):
        prev_outputs = state.prev_outputs
        unused_prev_outputs = set(prev_outputs.keys()).difference(state.used_output_names)
        for name in unused_prev_outputs:
            if name in state.cur_virtual or name in state.prev_persistent:
                output = Node(name)
                state.outputs.append(output)
                state.cur_used_virtual.add(name)
                state.used_output_names.add(name)
                state.edges.append(BoundEdge(IdentityEdge(), [prev_outputs[name]], output))

    @staticmethod
    def _connect_prev_virtual(state: LayerConnectionState):
        cur_inputs = state.cur_inputs
        unused_input_names = set(cur_inputs.keys()).difference(state.used_input_names)
        for name in unused_input_names:
            if name in state.prev_virtual:
                input_node = Node(name)
                output = cur_inputs[name]
                state.inputs.append(input_node)
                state.used_input_names.add(name)
                state.prev_used_virtual.add(name)
                state.edges.append(BoundEdge(IdentityEdge(), [input_node], output))

            elif name not in state.cur_optional:
                raise DependencyError(f"Previous layer must contain '{name}' node.")

    @staticmethod
    def _merge_virtual_nodes(state: LayerConnectionState):
        unused_cur_virtual = state.cur_virtual - state.cur_used_virtual
        unused_prev_virtual = state.prev_virtual - state.prev_used_virtual
        return unused_cur_virtual & unused_prev_virtual

    @staticmethod
    def get_essential_input_names(inputs: Sequence[Node], outputs: Sequence[Node], edges: BoundEdges):
        check_for_duplicates(inputs)
        tree_node_map = TreeNode.from_edges(edges)
        inputs = [tree_node_map[x] for x in inputs]

        essential_input_names = {}
        for o in outputs:
            output = tree_node_map[o]
            counts = count_entries(inputs, output)
            input_names = [x.name for x in inputs if counts.get(x, 0)]
            essential_input_names[o] = input_names
        return essential_input_names


def is_reachable(inputs: TreeNodes, output: TreeNode):
    def reachable(x: TreeNode):
        if x.is_leaf:
            return x in inputs

        return all(map(reachable, x.parents))

    inputs = set(inputs)
    return reachable(output)


def normalize_inherit(value, outputs) -> Tuple[Set[str], bool]:
    if isinstance(value, str):
        value = [value]

    if isinstance(value, bool):
        valid = value
        value = AntiSet(set(outputs))
    elif isinstance(value, AntiSet):
        valid = True
    else:
        value = set(value)
        valid = all(isinstance(node_name, str) for node_name in value)

    return value, valid


class PipelineContext(Context):
    def __init__(self, previous: Context, current: Context):
        self.previous = previous
        self.current = current

    def reverse(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges) -> Tuple[Nodes, BoundEdges]:
        outputs, edges = self.current.reverse(inputs, outputs, edges)
        outputs, edges = self.previous.reverse(inputs, outputs, edges)
        return outputs, edges

    def update(self, mapping: dict) -> 'Context':
        return PipelineContext(self.previous.update(mapping), self.current.update(mapping))


class BagContext(Context):
    def __init__(self, inputs: Nodes, outputs: Nodes, inherit: Set[str]):
        self.inputs = inputs
        self.outputs = outputs
        self.inherit = inherit

    def reverse(self, inputs: Nodes, outputs: Nodes, edges: BoundEdges) -> Tuple[Nodes, BoundEdges]:
        edges = list(edges)
        outputs = node_to_dict(outputs)
        # backward transforms
        for node in self.inputs:
            name = node.name
            if name in outputs:
                edges.append(IdentityEdge().bind(outputs[name], node))

        # collect the actual outputs
        actual = []
        mapping = TreeNode.from_edges(edges)
        leaves = [mapping[node] for node in inputs]
        for node in self.outputs:
            if is_reachable(leaves, mapping[node]):
                actual.append(node)

        # add inheritance
        add = self.inherit - set(node_to_dict(actual))
        for name, node in outputs.items():
            name = node.name
            if name in add:
                out = Node(name)
                edges.append(IdentityEdge().bind(node, out))
                actual.append(out)

        return actual, edges

    def update(self, mapping: dict) -> 'Context':
        return BagContext(
            update_map(self.inputs, mapping),
            update_map(self.outputs, mapping),
            self.inherit,
        )



from abc import ABC, abstractmethod
from functools import partial
from typing import Union, Dict, Callable

from contextlib import suppress
from gzip import GzipFile
from pathlib import Path

import os
import json
import pickle
import numpy as np

from .storage import Storage


class SerializerError(Exception):
    pass


class Serializer(ABC):
    @abstractmethod
    def save(self, value, folder: Path):
        """ Saves the ``value`` to ``folder`` """

    @abstractmethod
    def load(self, folder: Path, storage: Storage):
        """ Loads the value from ``folder`` """

    @staticmethod
    def _load_file(storage: Storage, loader: Callable, path: Path, *args, **kwargs):
        """ Useful function for loading files from storage """
        # it's simpler to wait for old caches to expire, than to convert them
        if path.is_symlink():
            return loader(path, *args, **kwargs)

        with open(path, 'r') as key:
            return storage.read(loader, key.read(), *args, **kwargs)


class ChainSerializer(Serializer):
    def __init__(self, *serializers: Serializer):
        self.serializers = serializers

    def save(self, value, folder: Path):
        for serializer in self.serializers:
            with suppress(SerializerError):
                return serializer.save(value, folder)

        raise SerializerError(f'No serializer was able to save to {folder}.')

    def load(self, folder: Path, storage: Storage):
        for serializer in self.serializers:
            with suppress(SerializerError):
                return serializer.load(folder, storage)

        raise SerializerError(f'No serializer was able to load from {folder}.')


class JsonSerializer(Serializer):
    def save(self, value, folder: Path):
        try:
            value = json.dumps(value)
        except TypeError as e:
            raise SerializerError from e

        with open(folder / 'value.json', 'w') as file:
            file.write(value)

    def load(self, folder: Path, storage: Storage):
        paths = list(folder.iterdir())
        if len(paths) != 1:
            raise SerializerError

        path, = paths
        if path.name != 'value.json':
            raise SerializerError

        def loader(x):
            with open(x, 'r') as file:
                return json.load(file)

        return self._load_file(storage, loader, folder / 'value.json')


class PickleSerializer(Serializer):
    def save(self, value, folder):
        try:
            value = pickle.dumps(value)
        except TypeError as e:
            raise SerializerError from e

        with open(folder / 'value.pkl', 'wb') as file:
            file.write(value)

    def load(self, folder: Path, storage: Storage):
        paths = list(folder.iterdir())
        if len(paths) != 1:
            raise SerializerError

        path, = paths
        if path.name != 'value.pkl':
            raise SerializerError

        def loader(x):
            with open(x, 'rb') as file:
                return pickle.load(file)

        return self._load_file(storage, loader, folder / 'value.pkl')


class NumpySerializer(Serializer):
    def __init__(self, compression: Union[int, Dict[type, int]] = None):
        self.compression = compression

    def _choose_compression(self, value):
        if isinstance(self.compression, int) or self.compression is None:
            return self.compression

        if isinstance(self.compression, dict):
            for dtype in self.compression:
                if np.issubdtype(value.dtype, dtype):
                    return self.compression[dtype]

    def save(self, value, folder: Path):
        value = np.asarray(value)
        compression = self._choose_compression(value)
        if compression is not None:
            assert isinstance(compression, int)
            with GzipFile(folder / 'value.npy.gz', 'wb', compresslevel=compression, mtime=0) as file:
                np.save(file, value)

        else:
            np.save(folder / 'value.npy', value)

    def load(self, folder: Path, storage: Storage):
        paths = list(folder.iterdir())
        if len(paths) != 1:
            raise SerializerError

        path, = paths
        if path.name == 'value.npy':
            loader = partial(np.load, allow_pickle=True)
        elif path.name == 'value.npy.gz':
            def loader(x):
                with GzipFile(x, 'rb') as file:
                    return np.load(file, allow_pickle=True)
        else:
            raise SerializerError

        return self._load_file(storage, loader, path)


class DictSerializer(Serializer):
    def __init__(self, serializer: Serializer):
        self.keys_filename = 'dict_keys.json'
        self.serializer = serializer

    def save(self, data: dict, folder: Path):
        if not isinstance(data, dict):
            raise SerializerError

        # TODO: remove all if at least one iteration fails
        keys_to_folder = {}
        for sub_folder, (key, value) in enumerate(data.items()):
            keys_to_folder[sub_folder] = key
            os.makedirs(folder / str(sub_folder), exist_ok=True)
            self.serializer.save(value, folder / str(sub_folder))

        with open(folder / self.keys_filename, 'w+') as f:
            json.dump(keys_to_folder, f)

    def load(self, folder: Path, storage: Storage):
        keys = folder / self.keys_filename
        if not keys.exists():
            raise SerializerError

        def loader(x):
            with open(x, 'r') as f:
                return json.load(f)

        keys_map = self._load_file(storage, loader, keys)
        data = {}
        for sub_folder, key in keys_map.items():
            data[key] = self.serializer.load(folder / sub_folder, storage)
        return data



from datetime import datetime
import gzip
import logging
import os
import shutil
import warnings
from pathlib import Path
from typing import Any, Union, Set

from ...exceptions import StorageCorruption
from ...storage import Storage
from ...storage.digest import digest_to_relative, get_digest_size
from ...serializers import Serializer
from ...storage.disk import DiskBase, copy_file
from ...storage.interface import QueryError, Key
from ...storage.utils import touch, create_folders, to_read_only, Reason
from ..compat import BadGzipFile

logger = logging.getLogger(__name__)

DATA_FOLDER = 'data'
TEMP_FOLDER = 'temp'
HASH_FILENAME = 'hash.bin'
TIME_FILENAME = 'time'
GZIP_COMPRESSION = 1


class CacheIndexStorage(DiskBase):
    def __init__(self, root: Path, storage: Storage, serializer: Serializer):
        super().__init__(root)
        self.storage = storage
        self.serializer = serializer

    def _check_consistency(self, base: Path, key: Key, value: Any, context):
        check_consistency(base / HASH_FILENAME, context, check_existence=True)

    def _write(self, base: Path, key: Key, value: Any, context: Any):
        data_folder, temp_folder = base / DATA_FOLDER, base / TEMP_FOLDER
        create_folders(data_folder, self.permissions, self.group)
        create_folders(temp_folder, self.permissions, self.group)

        self.serializer.save(value, temp_folder)
        self._mirror_to_storage(temp_folder, data_folder)
        self._save_meta(base, context)

    def _replicate(self, base: Path, key: Key, source: Path, context):
        # data
        shutil.copytree(source / DATA_FOLDER, base / DATA_FOLDER)
        # meta
        copy_file(source / HASH_FILENAME, base / HASH_FILENAME)
        copy_file(source / TIME_FILENAME, base / TIME_FILENAME)
        to_read_only(base / HASH_FILENAME, self.permissions, self.group)

    def _save_meta(self, base, pickled):
        hash_path, time_path = base / HASH_FILENAME, base / TIME_FILENAME
        # time
        with open(time_path, 'w'):
            pass
        os.chmod(time_path, 0o777)
        shutil.chown(time_path, group=self.group)
        # hash
        with gzip.GzipFile(hash_path, 'wb', compresslevel=GZIP_COMPRESSION, mtime=0) as file:
            file.write(pickled)
        to_read_only(hash_path, self.permissions, self.group)

    def _mirror_to_storage(self, source: Path, destination: Path):
        for file in source.glob('**/*'):
            target = destination / file.relative_to(source)
            if file.is_dir():
                target.mkdir(parents=True)

            else:
                with open(target, 'w') as fd:
                    fd.write(self.storage.write(file))
                os.remove(file)
                to_read_only(target, self.permissions, self.group)

        shutil.rmtree(source)

    def _cleanup_corrupted(self, folder, digest):
        warnings.warn(f'Corrupted storage at {self.root} for key {digest}. Cleaning up.', RuntimeWarning)
        shutil.rmtree(folder)

    def read(self, key, context):
        with self.locker.read(key):
            base = self.root / digest_to_relative(key, self.levels)
            if not base.exists():
                return None, False

            hash_path, time_path = base / HASH_FILENAME, base / TIME_FILENAME
            # we either have a valid folder
            if hash_path.exists() and time_path.exists():
                check_consistency(hash_path, context)
                touch(time_path)
                try:
                    # if couldn't find the hash - the cache is corrupted
                    return self.serializer.load(base / DATA_FOLDER, self.storage), True
                except QueryError:
                    pass

        # or it is corrupted, in which case we can remove it
        with self.locker.write(key):
            self._cleanup_corrupted(base, key)
            return None, False

    def inspect_entry(self, key: str, last_used: Union[float, datetime] = None) -> Union[Reason, Set[str]]:
        digest_size = get_digest_size(self.levels, string=True)
        if len(key) != digest_size:
            return Reason.WrongDigestSize

        data_key_size = self.storage.digest_size * 2
        base = self.root / digest_to_relative(key, self.levels)
        with self.locker.read(key):
            if {x.name for x in base.iterdir()} != {HASH_FILENAME, DATA_FOLDER, TIME_FILENAME}:
                return Reason.WrongFolderStructure

            if last_used is not None:
                if isinstance(last_used, datetime):
                    last_used = last_used.timestamp()
                if (base / TIME_FILENAME).stat().st_mtime < last_used:
                    return Reason.Expired

            try:
                with gzip.GzipFile(base / HASH_FILENAME, 'rb') as file:
                    hash_bytes = file.read()

            except BadGzipFile:
                return Reason.CorruptedHash

            real_digest = self.algorithm(hash_bytes).hexdigest()
            if key != real_digest:
                return Reason.WrongHash

            hashes = set()
            for file in (base / DATA_FOLDER).glob('**/*'):
                if file.is_dir():
                    continue

                if not file.is_file():
                    return Reason.CorruptedData

                with open(file, 'r') as content:
                    try:
                        data = content.read().strip()
                    except UnicodeDecodeError:
                        return Reason.CorruptedData

                    if len(data) != data_key_size:
                        return Reason.CorruptedData

                    hashes.add(data)

            return hashes


def check_consistency(hash_path, pickled, check_existence: bool = False):
    suggestion = f'You may want to delete the {hash_path.parent} folder.'
    if check_existence and not hash_path.exists():
        raise StorageCorruption(f'The pickled graph is missing. {suggestion}')
    try:
        with gzip.GzipFile(hash_path, 'rb') as file:
            dumped = file.read()
            if dumped != pickled:
                raise StorageCorruption(
                    f'The dumped and current pickle do not match at {hash_path}: {dumped} {pickled}. {suggestion}'
                )
    except BadGzipFile:
        raise StorageCorruption(f'The hash is corrupted. {suggestion}') from None



from .base import Container, EdgesBag


class PipelineContainer(EdgesBag):
    def __init__(self, head: EdgesBag, *tail: Container):
        self.containers = [head, *tail]
        for layer in tail:
            head = layer.wrap(head)

        state = head.freeze()
        super().__init__(state.inputs, state.outputs, state.edges, state.context, virtual_nodes=state.virtual_nodes,
                         persistent_nodes=state.persistent_nodes)

    def wrap(self, container: EdgesBag) -> EdgesBag:
        return PipelineContainer(container, *self.containers)


class LazyPipelineContainer(Container):
    def __init__(self, *containers: Container):
        self.containers = containers

    def wrap(self, container: EdgesBag) -> EdgesBag:
        return PipelineContainer(container, *self.containers)



import operator
import weakref
from pathlib import Path
from typing import Union, Sequence, Callable, Iterable, NamedTuple

import numpy as np

from .base import BaseLayer, CallableLayer
from ..cache.disk.index import CacheIndexStorage
from ..containers.cache import MemoryCacheContainer, DiskCacheContainer, CacheColumnsContainer
from ..containers.debug import HashDigestContainer
from ..containers.filter import FilterContainer
from ..containers.goup import GroupContainer, MultiGroupLayer
from ..containers.join import JoinContainer
from ..containers.merge import SwitchContainer
from ..containers.transform import TransformContainer
from ..engine.base import Node
from ..engine.edges import FunctionEdge
from ..serializers import Serializer, ChainSerializer, JsonSerializer, NumpySerializer, PickleSerializer
from ..storage import Storage, Disk
from ..storage.config import init_storage
from ..storage.interface import RemoteStorage
from ..utils import PathLike, StringsLike, AntiSet
from .utils import format_arguments


class Merge(CallableLayer):
    def __init__(self, *layers: CallableLayer):
        properties = [set(layer._properties) for layer in layers]
        inter = set.intersection(*properties)
        union = set.union(*properties)
        if inter != union:
            raise ValueError(f'All inputs must have the same properties: {properties}')
        properties = inter
        if not properties:
            raise ValueError('The datasets do not contain properties.')
        if len(properties) > 1:
            raise ValueError(f'Can\'t decide which property to use as ids.')
        ids_name, = properties

        id2dataset_index = {}
        for index, dataset in enumerate(layers):
            ids = getattr(dataset, ids_name)
            intersection = set(ids) & set(id2dataset_index)
            if intersection:
                raise RuntimeError(f'Ids {intersection} are duplicated in merged datasets.')

            id2dataset_index.update({i: index for i in ids})

        persistent = tuple(set.intersection(*(set(layer._container.persistent_nodes) for layer in layers)))
        super().__init__(SwitchContainer(
            id2dataset_index, [s._container for s in layers], ids_name, persistent),
            properties,
        )
        self._layers = layers

    def __repr__(self):
        return 'Merge' + format_arguments(self._layers)


class Join(CallableLayer):
    def __init__(self, left: CallableLayer, right: CallableLayer, on: StringsLike, pair_to_id: Callable):
        super().__init__(JoinContainer(left._container, right._container, to_seq(on), pair_to_id), ['ids'])


class Filter(BaseLayer[FilterContainer]):
    """
    Filters the `ids` of the current pipeline given a ``predicate``.

    Examples
    --------
    >>> dataset = Chain(
    >>>   source,  # dataset with `image` and `spacing` attributes
    >>>   Filter(lambda image, spacing: min(image.shape) > 30 and max(spacing) < 5),
    >>> )
    """

    def __init__(self, predicate: Callable, verbose: bool = False):
        super().__init__(FilterContainer(predicate, verbose))

    @classmethod
    def drop(cls, ids: Iterable[str], verbose: bool = False):
        """Removes the provided ``ids`` from the dataset."""
        assert not isinstance(ids, str)
        ids = tuple(sorted(set(ids)))
        assert all(isinstance(i, str) for i in ids)
        return cls(lambda id: id not in ids, verbose=verbose)

    @classmethod
    def keep(cls, ids: Iterable[str], verbose: bool = False):
        """Removes all the ids not present in ``ids``."""
        assert not isinstance(ids, str)
        ids = tuple(sorted(set(ids)))
        assert all(isinstance(i, str) for i in ids)
        return cls(lambda id: id in ids, verbose=verbose)

    def __repr__(self):
        args = ', '.join(self._container.names)
        return f'Filter({args})'


class GroupBy(BaseLayer[GroupContainer]):
    def __init__(self, name: str):
        super().__init__(GroupContainer(name))

    @staticmethod
    def _multiple(*names, **comparators):
        assert set(comparators).issubset(names)
        for name in names:
            comparators.setdefault(name, operator.eq)
        return BaseLayer(MultiGroupLayer(comparators))

    def __repr__(self):
        return f'GroupBy({repr(self._container.name)})'


class Apply(CallableLayer):
    def __init__(self, **transform: Callable):
        self.names = sorted(transform)

        inputs, outputs, edges = [], [], []
        for name, func in transform.items():
            inp, out = Node(name), Node(name)
            inputs.append(inp)
            outputs.append(out)
            edges.append(FunctionEdge(func, arity=1).bind(inp, out))

        super().__init__(TransformContainer(
            inputs, outputs, edges, forward_virtual=AntiSet(transform), backward_virtual=AntiSet()
        ), ())

    def __repr__(self):
        args = ', '.join(self.names)
        return f'Apply({args})'


def to_seq(x):
    if isinstance(x, str):
        x = [x]
    return x


def _resolve_serializer(serializer):
    if not isinstance(serializer, Serializer):
        serializer = ChainSerializer(*serializer)
    return serializer


class CacheLayer(BaseLayer):
    def __repr__(self):
        return self.__class__.__name__


class CacheToRam(CacheLayer):
    """
    Caches the fields from ``names`` to RAM.

    If ``size`` is not None - a LRU cache is used.
    """

    def __init__(self, names: StringsLike = None, *, size: int = None, impure: bool = False):
        self._cache_instances = weakref.WeakSet()
        super().__init__(MemoryCacheContainer(names, size, impure, self._cache_instances))

    def _clear(self):
        """ Clears all the values cached by this layer. """
        for cache in self._cache_instances:
            cache.clear()


class CacheIndex(NamedTuple):
    local: Sequence[PathLike]
    remote: Sequence[RemoteStorage] = ()


PathLikes = Union[PathLike, Sequence[PathLike]]
RemoteStorageLike = Union[RemoteStorage, Sequence[RemoteStorage]]
SerializersLike = Union[Serializer, Sequence[Serializer]]


class CacheToDisk(CacheLayer):
    """
    A persistent cache stored on disk.

    Parameters
    ----------
    index
        the folder(s) where the cache index is stored
    storage
        the storage which holds the actual data
    serializer
        the serializer used to save and load the data
    names
        field names that will be cached
    impure
        whether to allow caching of `impure` functions
    remote
        remote locations that are used to fetch the cache from (if available)
    """

    def __init__(self, index: PathLikes, storage: Storage, serializer: SerializersLike, names: StringsLike, *,
                 impure: bool = False, remote: RemoteStorageLike = ()):
        names, local, remote = _normalize_disk_arguments(index, remote, names, serializer, storage)
        super().__init__(DiskCacheContainer(names, local, remote, impure))

    @classmethod
    def simple(cls, *names, root: PathLike, serializer: Union[Serializer, Sequence[Serializer]] = None):
        """
        A simple version of caching to disk with adequate default settings.

        Parameters
        ----------
        names:
            the field names to cache
        root:
            the folder where the cache will be stored
        serializer
            the serializer used to save and load the data
        """
        root = Path(root)
        root.mkdir(exist_ok=True, parents=True)
        children = set(root.iterdir())
        index = root / 'index'
        storage = root / 'storage'

        if not children:
            init_storage(index, algorithm={'name': 'sha256'}, levels=[1, 31])
            init_storage(storage, algorithm={'name': 'sha256'}, levels=[1, 31])
        elif children != {index, storage}:
            names = tuple(x.name for x in children)
            raise FileNotFoundError(
                f'The root is expected to contain the "index" and "storage" folders, but found {names}'
            )

        if serializer is None:
            serializer = ChainSerializer(
                JsonSerializer(),
                NumpySerializer({np.bool_: 1, np.int_: 1}),
                PickleSerializer(),
            )

        return cls(index, Storage([Disk(storage)]), serializer, names)


class CacheColumns(CacheLayer):
    """
    A combination of a persistent cache stored on disk and a memory cache.
    The entries are stored on disk in shards, which speeds up read/write operations for large numbers of small files.

    Parameters
    ----------
    index
        the folder(s) where the cache index is stored
    storage
        the storage which holds the actual data
    serializer
        the serializer used to save and load the data
    names
        field names that will be cached
    remote
        remote locations that are used to fetch the cache from (if available)
    verbose
        whether to show a progressbar during cache generation
    shard_size
        the size of a disk storage shard. If int - an absolute size value is used,
        if float - a portion relative to the dataset is used,
        if None - all the entries are grouped in a single shard
    """

    def __init__(self, index: PathLikes, storage: Storage, serializer: SerializersLike, names: StringsLike, *,
                 verbose: bool = False, shard_size: Union[int, float, None] = None, remote: RemoteStorageLike = ()):
        if shard_size == 1:
            raise ValueError(f'Shard size of 1 is ambiguous. Use None if you want to have a single shard')
        names, local, remote = _normalize_disk_arguments(index, remote, names, serializer, storage)
        super().__init__(CacheColumnsContainer(names, local, remote, verbose=verbose, shard_size=shard_size))


class HashDigest(BaseLayer):
    def __init__(self, names: StringsLike):
        super().__init__(HashDigestContainer(to_seq(names)))


def _normalize_disk_arguments(local, remote, names, serializer, storage):
    names = to_seq(names)
    serializer = _resolve_serializer(serializer)
    if isinstance(local, CacheIndex):
        local, remote = local.local, local.remote
    if isinstance(local, (str, Path)):
        local = local,
    if isinstance(remote, RemoteStorage):
        remote = remote,
    local = [CacheIndexStorage(root, storage, serializer) for root in local]
    return names, local, remote



from abc import ABC, abstractmethod
from enum import Enum
from typing import Sequence, Tuple, Union, NamedTuple, Optional, Any, Generator

from .node_hash import NodeHash, NodeHashes


class Command(Enum):
    ParentHash, CurrentHash, ParentValue, Payload, Await, Call = range(6)
    Send, Store, Item, ComputeHash, Evaluate, AwaitThunk, Tuple, Return = range(-8, 0)


HashOutput = Tuple[NodeHash, Any]
Request = Tuple  # [RequestType, Any, ...]
Response = Union[NodeHash, Any, Tuple[NodeHash, Any]]


class Edge(ABC):
    def __init__(self, arity: int):
        self.arity = arity

    @abstractmethod
    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        """ Computes the hash of the output given the input hashes. """

    @abstractmethod
    def evaluate(self) -> Generator[Request, Response, Any]:
        """ Computes the output value. """

    def hash_graph(self, inputs: NodeHashes) -> NodeHash:
        """ Propagates the graph's hash without any control flow. """
        assert len(inputs) == self.arity
        return self._hash_graph(inputs)

    @abstractmethod
    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        pass

    def bind(self, inputs: Union['Node', 'Nodes'], output: 'Node') -> 'BoundEdge':
        if isinstance(inputs, Node):
            inputs = [inputs]
        assert len(inputs) == self.arity, (len(inputs), self.arity)
        return BoundEdge(self, inputs, output)


class TreeNode:
    __slots__ = 'name', '_edge'

    def __init__(self, name: str, edge: Optional[Tuple[Edge, Sequence['TreeNode']]]):
        self.name, self._edge = name, edge

    @property
    def is_leaf(self):
        return self._edge is None

    @property
    def edge(self):
        return self._edge[0]

    @property
    def parents(self):
        return self._edge[1]

    @classmethod
    def from_edges(cls, edges: Sequence['BoundEdge']) -> dict:
        def update(node: Node):
            if node in mapping:
                return mapping[node]

            bridge = bridges.get(node)
            if bridge is not None:
                bridge = bridge.edge, tuple(update(x) for x in bridge.inputs)
            mapping[node] = new = cls(node.name, bridge)
            return new

        nodes = set()
        bridges = {}
        # each edge is represented by its output
        for edge in edges:
            # TODO: replace by exception
            assert edge.output not in bridges, edge
            bridges[edge.output] = edge
            nodes.add(edge.output)
            nodes.update(edge.inputs)

        mapping = {}
        for n in nodes:
            update(n)
        return mapping

    def __repr__(self):
        return f'<TreeNode: {self.name}>'


class Node:
    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return f'<Node: {self.name}>'


class BoundEdge(NamedTuple):
    edge: Edge
    inputs: 'Nodes'
    output: Node

    __iter__ = None


TreeNodes = Sequence[TreeNode]
Nodes = Sequence[Node]
BoundEdges = Sequence[BoundEdge]
Edges = Sequence[Edge]


class HashError(RuntimeError):
    pass



from typing import Sequence, Any, Generator

from ..engine.edges import ConstantEdge, IdentityEdge
from ..engine.base import Node, NodeHash, Edge, NodeHashes, HashOutput, Request, Response, Command
from ..engine.node_hash import MergeHash
from ..utils import node_to_dict
from .base import EdgesBag


class SwitchContainer(EdgesBag):
    def __init__(self, id_to_index: dict, layers: Sequence[EdgesBag], keys_name: str, persistent_names: Sequence[str]):
        inputs = []
        groups = []
        edges = []
        # gather parts
        for layer in layers:
            params = layer.freeze()
            if len(params.inputs) != 1:
                raise ValueError('Each layer must have exactly one input')
            inputs.append(params.inputs[0])
            groups.append(node_to_dict(params.outputs))
            edges.extend(params.edges)

        # validate inputs
        inp = [x.name for x in inputs]
        if len(set(inp)) != 1:
            raise ValueError(f'Layer inputs must have the same name: {inp}')

        # create the new input
        inp = Node(inp[0])
        for node in inputs:
            edges.append(IdentityEdge().bind(inp, node))

        # create new outputs
        outputs = []
        common_outputs = set.intersection(*map(set, groups)) - {keys_name}
        for name in common_outputs:
            node = Node(name)
            branches = [group[name] for group in groups]
            outputs.append(node)
            edges.append(SwitchEdge(id_to_index, len(layers)).bind([inp] + branches, node))

        # and the keys
        ids = Node(keys_name)
        outputs.append(ids)
        edges.append(ConstantEdge(tuple(sorted(id_to_index))).bind([], ids))

        super().__init__([inp], outputs, edges, context=None, persistent_nodes=set(persistent_names))
        self.layers = layers


class SwitchEdge(Edge):
    def __init__(self, id_to_index: dict, n_branches: int):
        super().__init__(arity=1 + n_branches)
        self.id_to_index = id_to_index

    def compute_hash(self) -> Generator[Request, Response, HashOutput]:
        key = yield Command.ParentValue, 0
        try:
            idx = self.id_to_index[key]
        except KeyError:
            raise ValueError(f'Identifier {key} not found.') from None

        value = yield Command.ParentHash, idx + 1
        return value, idx

    def evaluate(self) -> Generator[Request, Response, Any]:
        payload = yield Command.Payload,
        value = yield Command.ParentValue, payload + 1
        return value

    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        return MergeHash(*inputs)



"""
This module contains a relaxed but reproducible version of cloudpickle:
    1. any object, after being pickled, must produce the same output every time
    2. two different objects cannot have equal results after pickling
    3. we don't need to restore the pickled object, so we can drop any information
        as long as it helps to achieve (or doesn't impede) 1 and 2
"""
import abc
import itertools
import pickle
import pickletools
import types
from operator import itemgetter
from typing import NamedTuple, Any
from contextlib import suppress
from enum import Enum
from io import BytesIO

from cloudpickle.cloudpickle import is_tornado_coroutine, PYPY, builtin_code_type, \
    _rebuild_tornado_coroutine, _find_imported_submodules, _BUILTIN_TYPE_NAMES, _builtin_type, _extract_class_dict

from .compat import DISPATCH, extract_func_data, Pickler, get_pickle_mode, PickleMode


def sort_dict(d):
    return tuple(sorted(d.items()))


VERSION_METHOD = '__getversion__'


class VersionedClass(NamedTuple):
    type: type
    version: Any


class PickleError(TypeError):
    pass


# new invalidation bugs will inevitably arise
# versioning will help diminish the pain from transitioning between updates
AVAILABLE_VERSIONS = 0,
*PREVIOUS_VERSIONS, LATEST_VERSION = AVAILABLE_VERSIONS

_custom = types.CodeType, types.FunctionType, type, property
for _key in _custom:
    DISPATCH.pop(_key, None)


class PortablePickler(Pickler):
    dispatch = DISPATCH

    def __init__(self, file, protocol=None, version=None):
        if version is None:
            version = LATEST_VERSION

        super().__init__(file, protocol=protocol)
        self.version = version
        self.dispatch_table = {
            types.CodeType: self.reduce_code,
            property: self.reduce_property,
        }

    def save(self, obj, *args, **kwargs):
        try:
            return super().save(obj, *args, **kwargs)
        except PickleError as e:
            raise PickleError(str(e)) from None
        except BaseException as e:
            raise PickleError(f'Exception "{e.__class__.__name__}: {e}" '
                              f'while pickling object {obj}') from None

    # copied from cloudpickle==2.0.0
    # used to save functions and classes
    def _save_reduce(self, func, args, state=None, listitems=None, dictitems=None, state_setter=None, obj=None):
        self.save_reduce(func, args, state=None, listitems=listitems, dictitems=dictitems, obj=obj)
        self.save(state_setter)
        self.save(obj)
        self.save(state)
        self.write(pickle.TUPLE2)
        self.write(pickle.REDUCE)
        self.write(pickle.POP)

    @staticmethod
    def _is_global(obj, name):
        return get_pickle_mode(obj, name) == PickleMode.Global

    @staticmethod
    def reduce_property(obj: property):
        return property, (obj.fget, obj.fset, obj.fdel, obj.__doc__)

    @staticmethod
    def reduce_code(obj: types.CodeType):
        """
        Same reducer as in cloudpickle, except `co_filename`, `co_firstlineno`, `lnotab` are not saved.
        """
        consts = obj.co_consts
        # remove the docstring:
        #  as of py3.9 the docstring is always stored in co_consts[0]
        #  associated issue: https://bugs.python.org/issue36521
        if consts and isinstance(consts[0], str):
            consts = list(consts)[1:]
            # TODO: this is not enough, None might be referenced in the bytecode under a wrong index
            if None in consts:
                consts.remove(None)
            consts = (None, *consts)

        if hasattr(obj, "co_posonlyargcount"):
            posonlyargcount = obj.co_posonlyargcount,
        else:
            posonlyargcount = ()

        args = (
            obj.co_argcount, *posonlyargcount,
            obj.co_kwonlyargcount, obj.co_nlocals, obj.co_stacksize,
            obj.co_flags, obj.co_code, consts, obj.co_names,
            obj.co_varnames,  # obj.co_filename,
            obj.co_name,  # obj.co_firstlineno, obj.co_lnotab,
            obj.co_freevars, obj.co_cellvars
        )
        return types.CodeType, args

    # function reducers will be used after we migrate to newer version
    def reduce_function(self, obj):
        """ Patched version that knows about __development__ mode """
        if self._is_global(obj, None):
            return NotImplemented
        elif PYPY and isinstance(obj.__code__, builtin_code_type):
            raise NotImplementedError
        else:
            return self.reduce_dynamic_function(obj)

    @staticmethod
    def reduce_dynamic_function(func):
        if is_tornado_coroutine(func):
            return _rebuild_tornado_coroutine, (func.__wrapped__,)

        # stuff we're dropping:
        #  1. base globals - only needed for unpickling
        #  2. __qualname__, __module__ - only used for debug
        #  3. __annotations__, __doc__ - not used at runtime

        code, f_globals, defaults, closure_values, dct = extract_func_data(func)
        f_globals, dct = map(sort_dict, [f_globals, dct])

        # args
        args = code, (len(closure_values) if closure_values is not None else -1)
        # state
        # TODO: do we need this?
        submodules = _find_imported_submodules(
            code,
            # same as f_globals.values()
            itertools.chain(map(itemgetter(1), f_globals), closure_values or ()),
        )
        name = func.__name__
        kw_defaults = getattr(func, '__kwdefaults__', None)
        if kw_defaults is not None:
            kw_defaults = sort_dict(kw_defaults)
        slot_state = f_globals, defaults, dct, closure_values, name, submodules, kw_defaults
        return types.FunctionType, args, (slot_state, func.__dict__)

    # for now we'll use the old `save_function`
    def save_function(self, obj, name=None):
        """ Patched version that knows about __development__ mode """
        if self._is_global(obj, name):
            return Pickler.save_global(self, obj, name=name)
        elif PYPY and isinstance(obj.__code__, builtin_code_type):
            return self.save_pypy_builtin_func(obj)
        else:
            return self._save_reduce(*self.reduce_dynamic_function(obj), obj=obj)

    dispatch[types.FunctionType] = save_function

    @staticmethod
    def _get_cls_params(obj):
        clsdict = _extract_class_dict(obj)
        clsdict.pop('__weakref__', None)

        if "_abc_impl" in clsdict:
            (registry, _, _, _) = abc._get_dump(obj)
            clsdict["_abc_impl"] = [subclass_weakref() for subclass_weakref in registry]

        # originally here was the __doc__
        type_kwargs = {}
        if hasattr(obj, "__slots__"):
            type_kwargs['__slots__'] = obj.__slots__
            if isinstance(obj.__slots__, str):
                clsdict.pop(obj.__slots__)
            else:
                for k in obj.__slots__:
                    clsdict.pop(k, None)

        __dict__ = clsdict.pop('__dict__', None)
        if isinstance(__dict__, property):
            type_kwargs['__dict__'] = __dict__

        # reproducibility
        clsdict.pop('__doc__', None)
        clsdict.pop('__module__', None)
        return clsdict, sort_dict(type_kwargs)

    def reduce_dynamic_class(self, obj):
        clsdict, type_kwargs = self._get_cls_params(obj)

        args = type(obj), obj.__name__, obj.__bases__, type_kwargs
        state = sort_dict(clsdict)
        return type, args, state

    def reduce_dynamic_enum(self, obj):
        clsdict, type_kwargs = self._get_cls_params(obj)

        members = {e.name: e.value for e in obj}
        for attrname in ["_generate_next_value_", "_member_names_", "_member_map_", "_member_type_",
                         "_value2member_map_"] + list(members):
            clsdict.pop(attrname, None)

        args = obj.__bases__, obj.__name__, sort_dict(members), obj.__module__
        state = sort_dict(clsdict)
        return type, args, state

    def save_dynamic_class(self, obj):
        method = self.reduce_dynamic_enum if isinstance(obj, Enum) else self.reduce_dynamic_class
        return self._save_reduce(*method(obj), obj=obj)

    def save_global(self, obj, name=None):
        """ Save a "global" which is not under __development__ """
        if obj is type(None):
            return self.save_reduce(type, (None,), obj=obj)
        elif obj is type(Ellipsis):
            return self.save_reduce(type, (Ellipsis,), obj=obj)
        elif obj is type(NotImplemented):
            return self.save_reduce(type, (NotImplemented,), obj=obj)
        elif obj in _BUILTIN_TYPE_NAMES:
            return self.save_reduce(_builtin_type, (_BUILTIN_TYPE_NAMES[obj],), obj=obj)
        elif name is not None:
            Pickler.save_global(self, obj, name=name)

        elif hasattr(obj, VERSION_METHOD):
            self.save(VersionedClass)
            version = getattr(obj, VERSION_METHOD)()
            Pickler.save_global(self, obj)
            self.save(version)
            self.write(pickle.TUPLE2)
            self.write(pickle.REDUCE)

        elif not self._is_global(obj, name=name):
            self.save_dynamic_class(obj)
        else:
            Pickler.save_global(self, obj, name=name)

    dispatch[type] = save_global

    with suppress(ImportError):
        from _functools import _lru_cache_wrapper

        # caching should not affect pickling
        def save_lru_cache(self, obj):
            # lru_cache uses functools.wrap
            self.save(obj.__wrapped__)

        dispatch[_lru_cache_wrapper] = save_lru_cache


def dumps(obj, protocol: int = None, version: int = None) -> bytes:
    with BytesIO() as file:
        PortablePickler(file, protocol=protocol, version=version).dump(obj)
        result = file.getvalue()
        result = pickletools.optimize(result)
        return result



import hashlib
from functools import partial
from pathlib import Path
from typing import Union, Dict, Any, Tuple

import humanfriendly
from pydantic import BaseModel, Extra, validator
from yaml import safe_load, safe_dump

from .locker import Locker, DummyLocker
from .utils import mkdir
from ..utils import PathLike

STORAGE_CONFIG_NAME = 'config.yml'


class HashConfig(BaseModel):
    name: str
    kwargs: Dict[str, Any] = None

    @validator('kwargs', always=True)
    def normalize_kwargs(cls, v):
        if v is None:
            return {}
        return v

    def build(self):
        cls = getattr(hashlib, self.name)
        if self.kwargs:
            cls = partial(cls, **self.kwargs)
        return cls


class LockerConfig(BaseModel):
    name: str
    args: Tuple = ()
    kwargs: Dict[str, Any] = None

    @validator('name')
    def name_exists(cls, name):
        for kls in Locker.__subclasses__():
            if kls.__name__ == name:
                return name

        raise ValueError(f'Could not find a locker named {name}')

    @validator('kwargs', always=True)
    def normalize_kwargs(cls, v):
        if v is None:
            return {}
        return v

    class Config:
        extra = Extra.forbid


class DiskConfig(BaseModel):
    hash: Union[str, HashConfig]
    levels: Tuple[int, ...]
    locker: Union[str, LockerConfig] = None
    free_disk_size: Union[int, str] = 0
    max_size: Union[int, str] = None

    @validator('free_disk_size', 'max_size')
    def to_size(cls, v):
        return parse_size(v)

    @validator('hash', pre=True)
    def normalize_hash(cls, v):
        if isinstance(v, str):
            v = {'name': v}
        return v

    @validator('locker', pre=True)
    def normalize_locker(cls, v):
        if isinstance(v, str):
            v = {'name': v}
        return v

    class Config:
        extra = Extra.forbid


def root_params(root: Path):
    stat = root.stat()
    return stat.st_mode & 0o777, stat.st_gid


def load_config(root: PathLike) -> DiskConfig:
    with open(Path(root) / STORAGE_CONFIG_NAME) as file:
        # TODO: assert read-only
        return DiskConfig(**safe_load(file))


def init_storage(root: PathLike, *, permissions: Union[int, None] = None, group: Union[str, int, None] = None,
                 algorithm: dict, levels: list, locker=None, exist_ok: bool = False, **params):
    root = Path(root)
    mkdir(root, permissions, group, parents=True, exist_ok=exist_ok)
    config = {
        'hash': algorithm,
        'levels': levels,
        **params,
    }
    if locker is not None:
        config['locker'] = locker

    with open(root / STORAGE_CONFIG_NAME, 'w') as file:
        safe_dump(config, file)


def make_locker(config: LockerConfig) -> Locker:
    if config is None:
        return DummyLocker()

    name = config.name
    for cls in Locker.__subclasses__():
        if cls.__name__ == name:
            return cls(*config.args, **config.kwargs)

    raise ValueError(f'Could not find a locker named {name}')


def parse_size(x):
    if isinstance(x, int):
        return x
    if isinstance(x, str):
        return humanfriendly.parse_size(x)
    if x is not None:
        raise ValueError(f"Couldn't understand the size format: {x}")



from .base import Chain, chained
from .metaclasses import Source, Transform, Mixin
from .blocks import CacheToRam, CacheToDisk, CacheColumns, Apply, Merge, Filter, GroupBy
from .decorators import *
from .edges import *
from .nodes import Input, Output, InverseInput, InverseOutput, Parameter
